"""
Support for actions and tasks.

Actions have a beginning and an eventual end, and can be nested. Tasks are
top-level actions.
"""

from __future__ import unicode_literals, absolute_import

import threading
from uuid import uuid4
from contextlib import contextmanager
from functools import partial
from inspect import getcallargs
from contextvars import ContextVar

from pyrsistent import field, PClass, optional, pmap_field, pvector
from boltons.funcutils import wraps
from six import text_type as unicode, PY3

from ._message import (
    WrittenMessage,
    EXCEPTION_FIELD,
    REASON_FIELD,
    TASK_UUID_FIELD,
    MESSAGE_TYPE_FIELD,
)
from ._util import safeunicode
from ._errors import _error_extraction

ACTION_STATUS_FIELD = "action_status"
ACTION_TYPE_FIELD = "action_type"

STARTED_STATUS = "started"
SUCCEEDED_STATUS = "succeeded"
FAILED_STATUS = "failed"

VALID_STATUSES = (STARTED_STATUS, SUCCEEDED_STATUS, FAILED_STATUS)

_ACTION_CONTEXT = ContextVar("eliot.action")

from ._message import TIMESTAMP_FIELD, TASK_LEVEL_FIELD


def current_action():
    """
    @return: The current C{Action} in context, or C{None} if none were set.
    """
    return _ACTION_CONTEXT.get(None)


class TaskLevel(object):
    """
    The location of a message within the tree of actions of a task.

    @ivar level: A pvector of integers. Each item indicates a child
        relationship, and the value indicates message count. E.g. C{[2,
        3]} indicates this is the third message within an action which is
        the second item in the task.
    """

    def __init__(self, level):
        self._level = level

    def as_list(self):
        """Return the current level.

        @return: List of integers.
        """
        return self._level[:]

    # Backwards compatibility:
    @property
    def level(self):
        return pvector(self._level)

    def __lt__(self, other):
        return self._level < other._level

    def __le__(self, other):
        return self._level <= other._level

    def __gt__(self, other):
        return self._level > other._level

    def __ge__(self, other):
        return self._level >= other._level

    def __eq__(self, other):
        if other.__class__ != TaskLevel:
            return False
        return self._level == other._level

    def __ne__(self, other):
        if other.__class__ != TaskLevel:
            return True
        return self._level != other._level

    def __hash__(self):
        return hash(tuple(self._level))

    @classmethod
    def fromString(cls, string):
        """
        Convert a serialized Unicode string to a L{TaskLevel}.

        @param string: Output of L{TaskLevel.toString}.

        @return: L{TaskLevel} parsed from the string.
        """
        return cls(level=[int(i) for i in string.split("/") if i])

    def toString(self):
        """
        Convert to a Unicode string, for serialization purposes.

        @return: L{unicode} representation of the L{TaskLevel}.
        """
        return "/" + "/".join(map(unicode, self._level))

    def next_sibling(self):
        """
        Return the next L{TaskLevel}, that is a task at the same level as this
        one, but one after.

        @return: L{TaskLevel} which follows this one.
        """
        new_level = self._level[:]
        new_level[-1] += 1
        return TaskLevel(level=new_level)

    def child(self):
        """
        Return a child of this L{TaskLevel}.

        @return: L{TaskLevel} which is the first child of this one.
        """
        new_level = self._level[:]
        new_level.append(1)
        return TaskLevel(level=new_level)

    def parent(self):
        """
        Return the parent of this L{TaskLevel}, or C{None} if it doesn't have
        one.

        @return: L{TaskLevel} which is the parent of this one.
        """
        if not self._level:
            return None
        return TaskLevel(level=self._level[:-1])

    def is_sibling_of(self, task_level):
        """
        Is this task a sibling of C{task_level}?
        """
        return self.parent() == task_level.parent()

    # PEP 8 compatibility:
    from_string = fromString
    to_string = toString


_TASK_ID_NOT_SUPPLIED = object()

import time


class Action(object):
    """
    Part of a nested heirarchy of ongoing actions.

    An action has a start and an end; a message is logged for each.

    Actions should only be used from a single thread, by implication the
    thread where they were created.

    @ivar _identification: Fields identifying this action.

    @ivar _successFields: Fields to be included in successful finish message.

    @ivar _finished: L{True} if the L{Action} has finished, otherwise L{False}.
    """

    def __init__(self, logger, task_uuid, task_level, action_type, serializers=None):
        """
        Initialize the L{Action} and log the start message.

        You probably do not want to use this API directly: use L{start_action}
        or L{startTask} instead.

        @param logger: The L{eliot.ILogger} to which to write
            messages.

        @param task_uuid: The uuid of the top-level task, e.g. C{"123525"}.

        @param task_level: The action's level in the task.
        @type task_level: L{TaskLevel}

        @param action_type: The type of the action,
            e.g. C{"yourapp:subsystem:dosomething"}.

        @param serializers: Either a L{eliot._validation._ActionSerializers}
            instance or C{None}. In the latter case no validation or
            serialization will be done for messages generated by the
            L{Action}.
        """
        self._successFields = {}
        self._logger = _output._DEFAULT_LOGGER if (logger is None) else logger
        self._task_level = task_level
        self._last_child = None
        self._identification = {
            TASK_UUID_FIELD: task_uuid,
            ACTION_TYPE_FIELD: action_type,
        }
        self._serializers = serializers
        self._finished = False

    @property
    def task_uuid(self):
        """
        @return str: the current action's task UUID.
        """
        return self._identification[TASK_UUID_FIELD]

    def serialize_task_id(self):
        """
        Create a unique identifier for the current location within the task.

        The format is C{b"<task_uuid>@<task_level>"}.

        @return: L{bytes} encoding the current location within the task.
        """
        return "{}@{}".format(
            self._identification[TASK_UUID_FIELD], self._nextTaskLevel().toString()
        ).encode("ascii")

    @classmethod
    def continue_task(cls, logger=None, task_id=_TASK_ID_NOT_SUPPLIED):
        """
        Start a new action which is part of a serialized task.

        @param logger: The L{eliot.ILogger} to which to write
            messages, or C{None} if the default one should be used.

        @param task_id: A serialized task identifier, the output of
            L{Action.serialize_task_id}, either ASCII-encoded bytes or unicode
            string. Required.

        @return: The new L{Action} instance.
        """
        if task_id is _TASK_ID_NOT_SUPPLIED:
            raise RuntimeError("You must supply a task_id keyword argument.")
        if isinstance(task_id, bytes):
            task_id = task_id.decode("ascii")
        uuid, task_level = task_id.split("@")
        action = cls(
            logger, uuid, TaskLevel.fromString(task_level), "eliot:remote_task"
        )
        action._start({})
        return action

    # Backwards-compat variants:
    serializeTaskId = serialize_task_id
    continueTask = continue_task

    def _nextTaskLevel(self):
        """
        Return the next C{task_level} for messages within this action.

        Called whenever a message is logged within the context of an action.

        @return: The message's C{task_level}.
        """
        if not self._last_child:
            self._last_child = self._task_level.child()
        else:
            self._last_child = self._last_child.next_sibling()
        return self._last_child

    def _start(self, fields):
        """
        Log the start message.

        The action identification fields, and any additional given fields,
        will be logged.

        In general you shouldn't call this yourself, instead using a C{with}
        block or L{Action.finish}.
        """
        fields[ACTION_STATUS_FIELD] = STARTED_STATUS
        fields[TIMESTAMP_FIELD] = time.time()
        fields.update(self._identification)
        fields[TASK_LEVEL_FIELD] = self._nextTaskLevel().as_list()
        if self._serializers is None:
            serializer = None
        else:
            serializer = self._serializers.start
        self._logger.write(fields, serializer)

    def finish(self, exception=None):
        """
        Log the finish message.

        The action identification fields, and any additional given fields,
        will be logged.

        In general you shouldn't call this yourself, instead using a C{with}
        block or L{Action.finish}.

        @param exception: C{None}, in which case the fields added with
            L{Action.addSuccessFields} are used. Or an L{Exception}, in
            which case an C{"exception"} field is added with the given
            L{Exception} type and C{"reason"} with its contents.
        """
        if self._finished:
            return
        self._finished = True
        serializer = None
        if exception is None:
            fields = self._successFields
            fields[ACTION_STATUS_FIELD] = SUCCEEDED_STATUS
            if self._serializers is not None:
                serializer = self._serializers.success
        else:
            fields = _error_extraction.get_fields_for_exception(self._logger, exception)
            fields[EXCEPTION_FIELD] = "%s.%s" % (
                exception.__class__.__module__,
                exception.__class__.__name__,
            )
            fields[REASON_FIELD] = safeunicode(exception)
            fields[ACTION_STATUS_FIELD] = FAILED_STATUS
            if self._serializers is not None:
                serializer = self._serializers.failure

        fields[TIMESTAMP_FIELD] = time.time()
        fields.update(self._identification)
        fields[TASK_LEVEL_FIELD] = self._nextTaskLevel().as_list()
        self._logger.write(fields, serializer)

    def child(self, logger, action_type, serializers=None):
        """
        Create a child L{Action}.

        Rather than calling this directly, you can use L{start_action} to
        create child L{Action} using the execution context.

        @param logger: The L{eliot.ILogger} to which to write
            messages.

        @param action_type: The type of this action,
            e.g. C{"yourapp:subsystem:dosomething"}.

        @param serializers: Either a L{eliot._validation._ActionSerializers}
            instance or C{None}. In the latter case no validation or
            serialization will be done for messages generated by the
            L{Action}.
        """
        newLevel = self._nextTaskLevel()
        return self.__class__(
            logger,
            self._identification[TASK_UUID_FIELD],
            newLevel,
            action_type,
            serializers,
        )

    def run(self, f, *args, **kwargs):
        """
        Run the given function with this L{Action} as its execution context.
        """
        parent = _ACTION_CONTEXT.set(self)
        try:
            return f(*args, **kwargs)
        finally:
            _ACTION_CONTEXT.reset(parent)

    def addSuccessFields(self, **fields):
        """
        Add fields to be included in the result message when the action
        finishes successfully.

        @param fields: Additional fields to add to the result message.
        """
        self._successFields.update(fields)

    # PEP 8 variant:
    add_success_fields = addSuccessFields

    @contextmanager
    def context(self):
        """
        Create a context manager that ensures code runs within action's context.

        The action does NOT finish when the context is exited.
        """
        parent = _ACTION_CONTEXT.set(self)
        try:
            yield self
        finally:
            _ACTION_CONTEXT.reset(parent)

    # Python context manager implementation:
    def __enter__(self):
        """
        Push this action onto the execution context.
        """
        self._parent_token = _ACTION_CONTEXT.set(self)
        return self

    def __exit__(self, type, exception, traceback):
        """
        Pop this action off the execution context, log finish message.
        """
        _ACTION_CONTEXT.reset(self._parent_token)
        self._parent_token = None
        self.finish(exception)

    ## Message logging
    def log(self, message_type, **fields):
        """Log individual message."""
        fields[TIMESTAMP_FIELD] = time.time()
        fields[TASK_UUID_FIELD] = self._identification[TASK_UUID_FIELD]
        fields[TASK_LEVEL_FIELD] = self._nextTaskLevel().as_list()
        fields[MESSAGE_TYPE_FIELD] = message_type
        self._logger.write(fields, fields.pop("__eliot_serializer__", None))


class WrongTask(Exception):
    """
    Tried to add a message to an action, but the message was from another
    task.
    """

    def __init__(self, action, message):
        Exception.__init__(
            self,
            "Tried to add {} to {}. Expected task_uuid = {}, got {}".format(
                message, action, action.task_uuid, message.task_uuid
            ),
        )


class WrongTaskLevel(Exception):
    """
    Tried to add a message to an action, but the task level of the message
    indicated that it was not a direct child.
    """

    def __init__(self, action, message):
        Exception.__init__(
            self,
            "Tried to add {} to {}, but {} is not a sibling of {}".format(
                message, action, message.task_level, action.task_level
            ),
        )


class WrongActionType(Exception):
    """
    Tried to end a message with a different action_type than the beginning.
    """

    def __init__(self, action, message):
        error_msg = "Tried to end {} with {}. Expected action_type = {}, got {}"
        Exception.__init__(
            self,
            error_msg.format(
                action,
                message,
                action.action_type,
                message.contents.get(ACTION_TYPE_FIELD, "<undefined>"),
            ),
        )


class InvalidStatus(Exception):
    """
    Tried to end a message with an invalid status.
    """

    def __init__(self, action, message):
        error_msg = "Tried to end {} with {}. Expected status {} or {}, got {}"
        Exception.__init__(
            self,
            error_msg.format(
                action,
                message,
                SUCCEEDED_STATUS,
                FAILED_STATUS,
                message.contents.get(ACTION_STATUS_FIELD, "<undefined>"),
            ),
        )


class DuplicateChild(Exception):
    """
    Tried to add a child to an action that already had a child at that task
    level.
    """

    def __init__(self, action, message):
        Exception.__init__(
            self,
            "Tried to add {} to {}, but already had child at {}".format(
                message, action, message.task_level
            ),
        )


class InvalidStartMessage(Exception):
    """
    Tried to start an action with an invalid message.
    """

    def __init__(self, message, reason):
        Exception.__init__(self, "Invalid start message {}: {}".format(message, reason))

    @classmethod
    def wrong_status(cls, message):
        return cls(message, 'must have status "STARTED"')

    @classmethod
    def wrong_task_level(cls, message):
        return cls(message, "first message must have task level ending in 1")


class WrittenAction(PClass):
    """
    An Action that has been logged.

    This class is intended to provide a definition within Eliot of what an
    action actually is, and a means of constructing actions that are known to
    be valid.

    @ivar WrittenMessage start_message: A start message whose task UUID and
        level match this action, or C{None} if it is not yet set on the
        action.
    @ivar WrittenMessage end_message: An end message hose task UUID and
        level match this action. Can be C{None} if the action is
        unfinished.
    @ivar TaskLevel task_level: The action's task level, e.g. if start
        message has level C{[2, 3, 1]} it will be
        C{TaskLevel(level=[2, 3])}.
    @ivar UUID task_uuid: The UUID of the task to which this action belongs.
    @ivar _children: A L{pmap} from L{TaskLevel} to the L{WrittenAction} and
        L{WrittenMessage} objects that make up this action.
    """

    start_message = field(type=optional(WrittenMessage), mandatory=True, initial=None)
    end_message = field(type=optional(WrittenMessage), mandatory=True, initial=None)
    task_level = field(type=TaskLevel, mandatory=True)
    task_uuid = field(type=unicode, mandatory=True, factory=unicode)
    # Pyrsistent doesn't support pmap_field with recursive types.
    _children = pmap_field(TaskLevel, object)

    @classmethod
    def from_messages(cls, start_message=None, children=pvector(), end_message=None):
        """
        Create a C{WrittenAction} from C{WrittenMessage}s and other
        C{WrittenAction}s.

        @param WrittenMessage start_message: A message that has
            C{ACTION_STATUS_FIELD}, C{ACTION_TYPE_FIELD}, and a C{task_level}
            that ends in C{1}, or C{None} if unavailable.
        @param children: An iterable of C{WrittenMessage} and C{WrittenAction}
        @param WrittenMessage end_message: A message that has the same
            C{action_type} as this action.

        @raise WrongTask: If C{end_message} has a C{task_uuid} that differs
            from C{start_message.task_uuid}.
        @raise WrongTaskLevel: If any child message or C{end_message} has a
            C{task_level} that means it is not a direct child.
        @raise WrongActionType: If C{end_message} has an C{ACTION_TYPE_FIELD}
            that differs from the C{ACTION_TYPE_FIELD} of C{start_message}.
        @raise InvalidStatus: If C{end_message} doesn't have an
            C{action_status}, or has one that is not C{SUCCEEDED_STATUS} or
            C{FAILED_STATUS}.
        @raise InvalidStartMessage: If C{start_message} does not have a
            C{ACTION_STATUS_FIELD} of C{STARTED_STATUS}, or if it has a
            C{task_level} indicating that it is not the first message of an
            action.

        @return: A new C{WrittenAction}.
        """
        actual_message = [
            message
            for message in [start_message, end_message] + list(children)
            if message
        ][0]
        action = cls(
            task_level=actual_message.task_level.parent(),
            task_uuid=actual_message.task_uuid,
        )
        if start_message:
            action = action._start(start_message)
        for child in children:
            if action._children.get(child.task_level, child) != child:
                raise DuplicateChild(action, child)
            action = action._add_child(child)
        if end_message:
            action = action._end(end_message)
        return action

    @property
    def action_type(self):
        """
        The type of this action, e.g. C{"yourapp:subsystem:dosomething"}.
        """
        if self.start_message:
            return self.start_message.contents[ACTION_TYPE_FIELD]
        elif self.end_message:
            return self.end_message.contents[ACTION_TYPE_FIELD]
        else:
            return None

    @property
    def status(self):
        """
        One of C{STARTED_STATUS}, C{SUCCEEDED_STATUS}, C{FAILED_STATUS} or
        C{None}.
        """
        message = self.end_message if self.end_message else self.start_message
        if message:
            return message.contents[ACTION_STATUS_FIELD]
        else:
            return None

    @property
    def start_time(self):
        """
        The Unix timestamp of when the action started, or C{None} if there has
        been no start message added so far.
        """
        if self.start_message:
            return self.start_message.timestamp

    @property
    def end_time(self):
        """
        The Unix timestamp of when the action ended, or C{None} if there has been
        no end message.
        """
        if self.end_message:
            return self.end_message.timestamp

    @property
    def exception(self):
        """
        If the action failed, the name of the exception that was raised to cause
        it to fail. If the action succeeded, or hasn't finished yet, then
        C{None}.
        """
        if self.end_message:
            return self.end_message.contents.get(EXCEPTION_FIELD, None)

    @property
    def reason(self):
        """
        The reason the action failed. If the action succeeded, or hasn't finished
        yet, then C{None}.
        """
        if self.end_message:
            return self.end_message.contents.get(REASON_FIELD, None)

    @property
    def children(self):
        """
        The list of child messages and actions sorted by task level, excluding the
        start and end messages.
        """
        return pvector(sorted(self._children.values(), key=lambda m: m.task_level))

    def _validate_message(self, message):
        """
        Is C{message} a valid direct child of this action?

        @param message: Either a C{WrittenAction} or a C{WrittenMessage}.

        @raise WrongTask: If C{message} has a C{task_uuid} that differs from the
            action's C{task_uuid}.
        @raise WrongTaskLevel: If C{message} has a C{task_level} that means
            it's not a direct child.
        """
        if message.task_uuid != self.task_uuid:
            raise WrongTask(self, message)
        if not message.task_level.parent() == self.task_level:
            raise WrongTaskLevel(self, message)

    def _add_child(self, message):
        """
        Return a new action with C{message} added as a child.

        Assumes C{message} is not an end message.

        @param message: Either a C{WrittenAction} or a C{WrittenMessage}.

        @raise WrongTask: If C{message} has a C{task_uuid} that differs from the
            action's C{task_uuid}.
        @raise WrongTaskLevel: If C{message} has a C{task_level} that means
            it's not a direct child.

        @return: A new C{WrittenAction}.
        """
        self._validate_message(message)
        level = message.task_level
        return self.transform(("_children", level), message)

    def _start(self, start_message):
        """
        Start this action given its start message.

        @param WrittenMessage start_message: A start message that has the
            same level as this action.

        @raise InvalidStartMessage: If C{start_message} does not have a
            C{ACTION_STATUS_FIELD} of C{STARTED_STATUS}, or if it has a
            C{task_level} indicating that it is not the first message of an
            action.
        """
        if start_message.contents.get(ACTION_STATUS_FIELD, None) != STARTED_STATUS:
            raise InvalidStartMessage.wrong_status(start_message)
        if start_message.task_level.level[-1] != 1:
            raise InvalidStartMessage.wrong_task_level(start_message)
        return self.set(start_message=start_message)

    def _end(self, end_message):
        """
        End this action with C{end_message}.

        Assumes that the action has not already been ended.

        @param WrittenMessage end_message: An end message that has the
            same level as this action.

        @raise WrongTask: If C{end_message} has a C{task_uuid} that differs
            from the action's C{task_uuid}.
        @raise WrongTaskLevel: If C{end_message} has a C{task_level} that means
            it's not a direct child.
        @raise InvalidStatus: If C{end_message} doesn't have an
            C{action_status}, or has one that is not C{SUCCEEDED_STATUS} or
            C{FAILED_STATUS}.

        @return: A new, completed C{WrittenAction}.
        """
        action_type = end_message.contents.get(ACTION_TYPE_FIELD, None)
        if self.action_type not in (None, action_type):
            raise WrongActionType(self, end_message)
        self._validate_message(end_message)
        status = end_message.contents.get(ACTION_STATUS_FIELD, None)
        if status not in (FAILED_STATUS, SUCCEEDED_STATUS):
            raise InvalidStatus(self, end_message)
        return self.set(end_message=end_message)


def start_action(logger=None, action_type="", _serializers=None, **fields):
    """
    Create a child L{Action}, figuring out the parent L{Action} from execution
    context, and log the start message.

    You can use the result as a Python context manager, or use the
    L{Action.finish} API to explicitly finish it.

         with start_action(logger, "yourapp:subsystem:dosomething",
                          entry=x) as action:
              do(x)
              result = something(x * 2)
              action.addSuccessFields(result=result)

    Or alternatively:

         action = start_action(logger, "yourapp:subsystem:dosomething",
                              entry=x)
         with action.context():
              do(x)
              result = something(x * 2)
              action.addSuccessFields(result=result)
         action.finish()

    @param logger: The L{eliot.ILogger} to which to write messages, or
        C{None} to use the default one.

    @param action_type: The type of this action,
        e.g. C{"yourapp:subsystem:dosomething"}.

    @param _serializers: Either a L{eliot._validation._ActionSerializers}
        instance or C{None}. In the latter case no validation or serialization
        will be done for messages generated by the L{Action}.

    @param fields: Additional fields to add to the start message.

    @return: A new L{Action}.
    """
    parent = current_action()
    if parent is None:
        return startTask(logger, action_type, _serializers, **fields)
    else:
        action = parent.child(logger, action_type, _serializers)
        action._start(fields)
        return action


def startTask(logger=None, action_type="", _serializers=None, **fields):
    """
    Like L{action}, but creates a new top-level L{Action} with no parent.

    @param logger: The L{eliot.ILogger} to which to write messages, or
        C{None} to use the default one.

    @param action_type: The type of this action,
        e.g. C{"yourapp:subsystem:dosomething"}.

    @param _serializers: Either a L{eliot._validation._ActionSerializers}
        instance or C{None}. In the latter case no validation or serialization
        will be done for messages generated by the L{Action}.

    @param fields: Additional fields to add to the start message.

    @return: A new L{Action}.
    """
    action = Action(
        logger, unicode(uuid4()), TaskLevel(level=[]), action_type, _serializers
    )
    action._start(fields)
    return action


class TooManyCalls(Exception):
    """
    The callable was called more than once.

    This typically indicates a coding bug: the result of
    C{preserve_context} should only be called once, and
    C{preserve_context} should therefore be called each time you want to
    pass the callable to a thread.
    """


def preserve_context(f):
    """
    Package up the given function with the current Eliot context, and then
    restore context and call given function when the resulting callable is
    run. This allows continuing the action context within a different thread.

    The result should only be used once, since it relies on
    L{Action.serialize_task_id} whose results should only be deserialized
    once.

    @param f: A callable.

    @return: One-time use callable that calls given function in context of
        a child of current Eliot action.
    """
    action = current_action()
    if action is None:
        return f
    task_id = action.serialize_task_id()
    called = threading.Lock()

    def restore_eliot_context(*args, **kwargs):
        # Make sure the function has not already been called:
        if not called.acquire(False):
            raise TooManyCalls(f)

        with Action.continue_task(task_id=task_id):
            return f(*args, **kwargs)

    return restore_eliot_context


def log_call(
    wrapped_function=None, action_type=None, include_args=None, include_result=True
):
    """Decorator/decorator factory that logs inputs and the return result.

    If used with inputs (i.e. as a decorator factory), it accepts the following
    parameters:

    @param action_type: The action type to use.  If not given the function name
        will be used.
    @param include_args: If given, should be a list of strings, the arguments to log.
    @param include_result: True by default. If False, the return result isn't logged.
    """
    if wrapped_function is None:
        return partial(
            log_call,
            action_type=action_type,
            include_args=include_args,
            include_result=include_result,
        )

    if action_type is None:
        if PY3:
            action_type = "{}.{}".format(
                wrapped_function.__module__, wrapped_function.__qualname__
            )
        else:
            action_type = wrapped_function.__name__

    if PY3 and include_args is not None:
        from inspect import signature

        sig = signature(wrapped_function)
        if set(include_args) - set(sig.parameters):
            raise ValueError(
                (
                    "include_args ({}) lists arguments not in the " "wrapped function"
                ).format(include_args)
            )

    @wraps(wrapped_function)
    def logging_wrapper(*args, **kwargs):
        callargs = getcallargs(wrapped_function, *args, **kwargs)

        # Remove self is it's included:
        if "self" in callargs:
            callargs.pop("self")

        # Filter arguments to log, if necessary:
        if include_args is not None:
            callargs = {k: callargs[k] for k in include_args}

        with start_action(action_type=action_type, **callargs) as ctx:
            result = wrapped_function(*args, **kwargs)
            if include_result:
                ctx.add_success_fields(result=result)
            return result

    return logging_wrapper


def log_message(message_type, **fields):
    """Log a message in the context of the current action.

    If there is no current action, a new UUID will be generated.
    """
    # Loggers will hopefully go away...
    logger = fields.pop("__eliot_logger__", None)
    action = current_action()
    if action is None:
        action = Action(logger, str(uuid4()), TaskLevel(level=[]), "")
    action.log(message_type, **fields)


from . import _output



"""
Error-handling utility code.
"""

from __future__ import unicode_literals

from inspect import getmro


class ErrorExtraction(object):
    """
    Extract fields from exceptions for failed-action messages.

    @ivar registry: Map exception class to function that extracts fields.
    """

    def __init__(self):
        self.registry = {}

    def register_exception_extractor(self, exception_class, extractor):
        """
        Register a function that converts exceptions to fields.

        @param exception_class: Class to register for.

        @param extractor: Single-argument callable that takes an exception
            of the given class (or a subclass) and returns a dictionary,
            fields to include in a failed action message.
        """
        self.registry[exception_class] = extractor

    def get_fields_for_exception(self, logger, exception):
        """
        Given an exception instance, return fields to add to the failed action
        message.

        @param logger: ``ILogger`` currently being used.
        @param exception: An exception instance.

        @return: Dictionary with fields to include.
        """
        for klass in getmro(exception.__class__):
            if klass in self.registry:
                extractor = self.registry[klass]
                try:
                    return extractor(exception)
                except:
                    from ._traceback import write_traceback

                    write_traceback(logger)
                    return {}
        return {}


_error_extraction = ErrorExtraction()
register_exception_extractor = _error_extraction.register_exception_extractor
get_fields_for_exception = _error_extraction.get_fields_for_exception

# Default handler for OSError and IOError by registered EnvironmentError:
register_exception_extractor(EnvironmentError, lambda e: {"errno": e.errno})



"""Tests for eliot.dask."""

from unittest import TestCase, skipUnless

from ..testing import capture_logging, LoggedAction, LoggedMessage
from .. import start_action, log_message

try:
    import dask
    from dask.bag import from_sequence
    from dask.distributed import Client
    import dask.dataframe as dd
    import pandas as pd
except ImportError:
    dask = None
else:
    from ..dask import (
        compute_with_trace,
        _RunWithEliotContext,
        _add_logging,
        persist_with_trace,
    )


@skipUnless(dask, "Dask not available.")
class DaskTests(TestCase):
    """Tests for end-to-end functionality."""

    def setUp(self):
        dask.config.set(scheduler="threading")

    def test_compute(self):
        """compute_with_trace() runs the same logic as compute()."""
        bag = from_sequence([1, 2, 3])
        bag = bag.map(lambda x: x * 7).map(lambda x: x * 4)
        bag = bag.fold(lambda x, y: x + y)
        self.assertEqual(dask.compute(bag), compute_with_trace(bag))

    def test_future(self):
        """compute_with_trace() can handle Futures."""
        client = Client(processes=False)
        self.addCleanup(client.shutdown)
        [bag] = dask.persist(from_sequence([1, 2, 3]))
        bag = bag.map(lambda x: x * 5)
        result = dask.compute(bag)
        self.assertEqual(result, ([5, 10, 15],))
        self.assertEqual(result, compute_with_trace(bag))

    def test_persist_result(self):
        """persist_with_trace() runs the same logic as process()."""
        client = Client(processes=False)
        self.addCleanup(client.shutdown)
        bag = from_sequence([1, 2, 3])
        bag = bag.map(lambda x: x * 7)
        self.assertEqual(
            [b.compute() for b in dask.persist(bag)],
            [b.compute() for b in persist_with_trace(bag)],
        )

    def test_persist_pandas(self):
        """persist_with_trace() with a Pandas dataframe.

        This ensures we don't blow up, which used to be the case.
        """
        df = pd.DataFrame()
        df = dd.from_pandas(df, npartitions=1)
        persist_with_trace(df)

    @capture_logging(None)
    def test_persist_logging(self, logger):
        """persist_with_trace() preserves Eliot context."""

        def persister(bag):
            [bag] = persist_with_trace(bag)
            return dask.compute(bag)

        self.assert_logging(logger, persister, "dask:persist")

    @capture_logging(None)
    def test_compute_logging(self, logger):
        """compute_with_trace() preserves Eliot context."""
        self.assert_logging(logger, compute_with_trace, "dask:compute")

    def assert_logging(self, logger, run_with_trace, top_action_name):
        """Utility function for _with_trace() logging tests."""

        def mult(x):
            log_message(message_type="mult")
            return x * 4

        def summer(x, y):
            log_message(message_type="finally")
            return x + y

        bag = from_sequence([1, 2])
        bag = bag.map(mult).fold(summer)
        with start_action(action_type="act1"):
            run_with_trace(bag)

        [logged_action] = LoggedAction.ofType(logger.messages, "act1")
        self.assertEqual(
            logged_action.type_tree(),
            {
                "act1": [
                    {
                        top_action_name: [
                            {"eliot:remote_task": ["dask:task", "mult"]},
                            {"eliot:remote_task": ["dask:task", "mult"]},
                            {"eliot:remote_task": ["dask:task", "finally"]},
                        ]
                    }
                ]
            },
        )

        # Make sure dependencies are tracked:
        mult1_msg, mult2_msg, final_msg = LoggedMessage.ofType(
            logger.messages, "dask:task"
        )
        self.assertEqual(
            sorted(final_msg.message["dependencies"]),
            sorted([mult1_msg.message["key"], mult2_msg.message["key"]]),
        )

        # Make sure dependencies are logically earlier in the logs:
        self.assertTrue(
            mult1_msg.message["task_level"] < final_msg.message["task_level"]
        )
        self.assertTrue(
            mult2_msg.message["task_level"] < final_msg.message["task_level"]
        )


@skipUnless(dask, "Dask not available.")
class AddLoggingTests(TestCase):
    """Tests for _add_logging()."""

    maxDiff = None

    def test_add_logging_to_full_graph(self):
        """_add_logging() recreates Dask graph with wrappers."""
        bag = from_sequence([1, 2, 3])
        bag = bag.map(lambda x: x * 7).map(lambda x: x * 4)
        bag = bag.fold(lambda x, y: x + y)
        graph = bag.__dask_graph__()

        # Add logging:
        with start_action(action_type="bleh"):
            logging_added = _add_logging(graph)

        # Ensure resulting graph hasn't changed substantively:
        logging_removed = {}
        for key, value in logging_added.items():
            if callable(value[0]):
                func, args = value[0], value[1:]
                self.assertIsInstance(func, _RunWithEliotContext)
                value = (func.func,) + args
            logging_removed[key] = value

        self.assertEqual(logging_removed, graph)

    def test_add_logging_explicit(self):
        """_add_logging() on more edge cases of the graph."""

        def add(s):
            return s + "s"

        def add2(s):
            return s + "s"

        # b runs first, then d, then a and c.
        graph = {
            "a": "d",
            "d": [1, 2, (add, "b")],
            ("b", 0): 1,
            "c": (add2, "d"),
        }

        with start_action(action_type="bleh") as action:
            task_id = action.task_uuid
            self.assertEqual(
                _add_logging(graph),
                {
                    "d": [
                        1,
                        2,
                        (
                            _RunWithEliotContext(
                                task_id=task_id + "@/2",
                                func=add,
                                key="d",
                                dependencies=["b"],
                            ),
                            "b",
                        ),
                    ],
                    "a": "d",
                    ("b", 0): 1,
                    "c": (
                        _RunWithEliotContext(
                            task_id=task_id + "@/3",
                            func=add2,
                            key="c",
                            dependencies=["d"],
                        ),
                        "d",
                    ),
                },
            )



"""
Tests for L{eliot.logwriter}.
"""

from __future__ import unicode_literals

import time
import threading

# Make sure to use StringIO that only accepts unicode:
from io import BytesIO, StringIO
from unittest import skipIf
import json as pyjson
from warnings import catch_warnings, simplefilter

from six import PY2

try:
    from zope.interface.verify import verifyClass
    from twisted.internet import reactor
    from twisted.trial.unittest import TestCase
    from twisted.application.service import IService
    from twisted.python import threadable
except ImportError:
    # Make tests not run at all.
    TestCase = object
else:
    # Make sure we always import this if Twisted is available, so broken
    # logwriter.py causes a failure:
    from ..logwriter import ThreadedFileWriter, ThreadedWriter

from .. import Logger, removeDestination, FileDestination


class BlockingFile(object):
    """
    A file-like whose writes can be blocked.

    Also, allow calling C{getvalue} after C{close}, unlike L{BytesIO}.
    """

    def __init__(self):
        self.file = BytesIO()
        self.lock = threading.Lock()
        self.data = b""

    def block(self):
        """
        Prevent writes until L{unblock} is called.
        """
        self.lock.acquire()

    def unblock(self):
        """
        Allow writes if L{block} was previous called.
        """
        self.lock.release()

    def getvalue(self):
        """
        Get written bytes.

        @return: Written bytes.
        """
        return self.data

    def write(self, data):
        with self.lock:
            self.file.write(data)

    def flush(self):
        self.data = self.file.getvalue()

    def close(self):
        self.file.close()


class ThreadedWriterTests(TestCase):
    """
    Tests for L{ThreadedWriter}.

    Many of these tests involve interactions across threads, so they
    arbitrarily wait for up to 5 seconds to reduce chances of slow thread
    switching causing the test to fail.
    """

    def test_interface(self):
        """
        L{ThreadedWriter} provides L{IService}.
        """
        verifyClass(IService, ThreadedWriter)

    def test_name(self):
        """
        L{ThreadedWriter} has a name.
        """
        self.assertEqual(ThreadedWriter.name, "Eliot Log Writer")

    def test_startServiceRunning(self):
        """
        L{ThreadedWriter.startService} starts the service as required by the
        L{IService} interface.
        """
        writer = ThreadedWriter(FileDestination(file=BytesIO()), reactor)
        self.assertFalse(writer.running)
        writer.startService()
        self.addCleanup(writer.stopService)
        self.assertTrue(writer.running)

    def test_stopServiceRunning(self):
        """
        L{ThreadedWriter.stopService} stops the service as required by the
        L{IService} interface.
        """
        writer = ThreadedWriter(FileDestination(file=BytesIO()), reactor)
        writer.startService()
        d = writer.stopService()
        d.addCallback(lambda _: self.assertFalse(writer.running))
        return d

    def test_startServiceStartsThread(self):
        """
        L{ThreadedWriter.startService} starts up a thread running
        L{ThreadedWriter._writer}.
        """
        previousThreads = threading.enumerate()
        result = []
        event = threading.Event()

        def _writer():
            current = threading.currentThread()
            if current not in previousThreads:
                result.append(current)
            event.set()

        writer = ThreadedWriter(FileDestination(file=BytesIO()), reactor)
        writer._writer = _writer
        writer.startService()
        event.wait()
        self.assertTrue(result)
        # Make sure thread is dead so it doesn't die half way through another
        # test:
        result[0].join(5)

    def test_stopServiceStopsThread(self):
        """
        L{ThreadedWriter.stopService} stops the writer thread.
        """
        previousThreads = set(threading.enumerate())
        writer = ThreadedWriter(FileDestination(file=BytesIO()), reactor)
        writer.startService()
        start = time.time()
        while set(threading.enumerate()) == previousThreads and (
            time.time() - start < 5
        ):
            time.sleep(0.0001)
        # If not true the next assertion might pass by mistake:
        self.assertNotEqual(set(threading.enumerate()), previousThreads)
        writer.stopService()
        while set(threading.enumerate()) != previousThreads and (
            time.time() - start < 5
        ):
            time.sleep(0.0001)
        self.assertEqual(set(threading.enumerate()), previousThreads)

    def test_stopServiceFinishesWriting(self):
        """
        L{ThreadedWriter.stopService} stops the writer thread, but only after
        all queued writes are written out.
        """
        f = BlockingFile()
        writer = ThreadedWriter(FileDestination(file=f), reactor)
        f.block()
        writer.startService()
        for i in range(100):
            writer({"write": 123})
        threads = threading.enumerate()
        writer.stopService()
        # Make sure writes didn't happen before the stopService, thus making the
        # test pointless:
        self.assertEqual(f.getvalue(), b"")
        f.unblock()
        start = time.time()
        while threading.enumerate() == threads and time.time() - start < 5:
            time.sleep(0.0001)
        self.assertEqual(f.getvalue(), b'{"write": 123}\n' * 100)

    def test_stopServiceResult(self):
        """
        L{ThreadedWriter.stopService} returns a L{Deferred} that fires only
        after the thread has shut down.
        """
        f = BlockingFile()
        writer = ThreadedWriter(FileDestination(file=f), reactor)
        f.block()
        writer.startService()

        writer({"hello": 123})
        threads = threading.enumerate()
        d = writer.stopService()
        f.unblock()

        def done(_):
            self.assertEqual(f.getvalue(), b'{"hello": 123}\n')
            self.assertNotEqual(threading.enumerate(), threads)

        d.addCallback(done)
        return d

    def test_noChangeToIOThread(self):
        """
        Running a L{ThreadedWriter} doesn't modify the Twisted registered IO
        thread.
        """
        writer = ThreadedWriter(FileDestination(file=BytesIO()), reactor)
        writer.startService()
        d = writer.stopService()
        # Either the current thread (the one running the tests) is the the I/O
        # thread or the I/O thread was never set. Either may happen depending on
        # how and whether the reactor has been started by the unittesting
        # framework.
        d.addCallback(
            lambda _: self.assertIn(
                threadable.ioThread, (None, threading.currentThread().ident)
            )
        )
        return d

    def test_startServiceRegistersDestination(self):
        """
        L{ThreadedWriter.startService} registers itself as an Eliot log
        destination.
        """
        f = BlockingFile()
        writer = ThreadedWriter(FileDestination(file=f), reactor)
        writer.startService()
        Logger().write({"x": "abc"})
        d = writer.stopService()
        d.addCallback(lambda _: self.assertIn(b"abc", f.getvalue()))
        return d

    def test_stopServiceUnregistersDestination(self):
        """
        L{ThreadedWriter.stopService} unregisters itself as an Eliot log
        destination.
        """
        writer = ThreadedWriter(FileDestination(file=BytesIO()), reactor)
        writer.startService()
        d = writer.stopService()
        d.addCallback(lambda _: removeDestination(writer))
        return self.assertFailure(d, ValueError)

    def test_call(self):
        """
        The message passed to L{ThreadedWriter.__call__} is passed to the
        underlying destination in the writer thread.
        """
        result = []

        def destination(message):
            result.append((message, threading.currentThread().ident))

        writer = ThreadedWriter(destination, reactor)
        writer.startService()
        thread_ident = writer._thread.ident
        msg = {"key": 123}
        writer(msg)
        d = writer.stopService()
        d.addCallback(lambda _: self.assertEqual(result, [(msg, thread_ident)]))
        return d


class ThreadedFileWriterTests(TestCase):
    """
    Tests for ``ThreadedFileWriter``.
    """

    def test_deprecation_warning(self):
        """
        Instantiating ``ThreadedFileWriter`` gives a ``DeprecationWarning``.
        """
        with catch_warnings(record=True) as warnings:
            ThreadedFileWriter(BytesIO(), reactor)
            simplefilter("always")  # Catch all warnings
            self.assertEqual(warnings[-1].category, DeprecationWarning)

    def test_write(self):
        """
        Messages passed to L{ThreadedFileWriter.__call__} are then written by
        the writer thread with a newline added.
        """
        f = BytesIO()
        writer = ThreadedFileWriter(f, reactor)
        writer.startService()
        self.addCleanup(writer.stopService)

        writer({"hello": 123})
        start = time.time()
        while not f.getvalue() and time.time() - start < 5:
            time.sleep(0.0001)
        self.assertEqual(f.getvalue(), b'{"hello": 123}\n')

    @skipIf(PY2, "Python 2 files always accept bytes")
    def test_write_unicode(self):
        """
        Messages passed to L{ThreadedFileWriter.__call__} are then written by
        the writer thread with a newline added to files that accept
        unicode.
        """
        f = StringIO()
        writer = ThreadedFileWriter(f, reactor)
        writer.startService()
        self.addCleanup(writer.stopService)

        original = {"hello\u1234": 123}
        writer(original)
        start = time.time()
        while not f.getvalue() and time.time() - start < 5:
            time.sleep(0.0001)
        self.assertEqual(f.getvalue(), pyjson.dumps(original) + "\n")

    def test_stopServiceClosesFile(self):
        """
        L{ThreadedWriter.stopService} closes the file.
        """
        f = BytesIO()
        writer = ThreadedFileWriter(f, reactor)
        writer.startService()
        d = writer.stopService()

        def done(_):
            self.assertTrue(f.closed)

        d.addCallback(done)
        return d



"""
Tests for C{eliot.prettyprint}.
"""

from unittest import TestCase
from subprocess import check_output, Popen, PIPE
from collections import OrderedDict
from datetime import datetime

from pyrsistent import pmap

from .._bytesjson import dumps
from ..prettyprint import pretty_format, compact_format, REQUIRED_FIELDS

SIMPLE_MESSAGE = {
    "timestamp": 1443193754,
    "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
    "message_type": "messagey",
    "task_level": [1, 2],
    "keys": [123, 456],
}

UNTYPED_MESSAGE = {
    "timestamp": 1443193754,
    "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
    "task_level": [1],
    "key": 1234,
    "abc": "def",
}


class FormattingTests(TestCase):
    """
    Tests for L{pretty_format}.
    """

    def test_message(self):
        """
        A typed message is printed as expected.
        """
        self.assertEqual(
            pretty_format(SIMPLE_MESSAGE),
            """\
8c668cde-235b-4872-af4e-caea524bd1c0 -> /1/2
2015-09-25T15:09:14Z
  message_type: 'messagey'
  keys: [123, 456]
""",
        )

    def test_untyped_message(self):
        """
        A message with no type is printed as expected.
        """
        self.assertEqual(
            pretty_format(UNTYPED_MESSAGE),
            """\
8c668cde-235b-4872-af4e-caea524bd1c0 -> /1
2015-09-25T15:09:14Z
  abc: 'def'
  key: 1234
""",
        )

    def test_action(self):
        """
        An action message is printed as expected.
        """
        message = {
            "task_uuid": "8bc6ded2-446c-4b6d-abbc-4f21f1c9a7d8",
            "place": "Statue #1",
            "task_level": [2, 2, 2, 1],
            "action_type": "visited",
            "timestamp": 1443193958.0,
            "action_status": "started",
        }
        self.assertEqual(
            pretty_format(message),
            """\
8bc6ded2-446c-4b6d-abbc-4f21f1c9a7d8 -> /2/2/2/1
2015-09-25T15:12:38Z
  action_type: 'visited'
  action_status: 'started'
  place: 'Statue #1'
""",
        )

    def test_multi_line(self):
        """
        Multiple line values are indented nicely.
        """
        message = {
            "timestamp": 1443193754,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
            "key": "hello\nthere\nmonkeys!\n",
            "more": "stuff",
        }
        self.assertEqual(
            pretty_format(message),
            """\
8c668cde-235b-4872-af4e-caea524bd1c0 -> /1
2015-09-25T15:09:14Z
  key: 'hello
     |  there
     |  monkeys!
     |  '
  more: 'stuff'
""",
        )

    def test_tabs(self):
        """
        Tabs are formatted as tabs, not quoted.
        """
        message = {
            "timestamp": 1443193754,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
            "key": "hello\tmonkeys!",
        }
        self.assertEqual(
            pretty_format(message),
            """\
8c668cde-235b-4872-af4e-caea524bd1c0 -> /1
2015-09-25T15:09:14Z
  key: 'hello	monkeys!'
""",
        )

    def test_structured(self):
        """
        Structured field values (e.g. a dictionary) are formatted in a helpful
        manner.
        """
        message = {
            "timestamp": 1443193754,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
            "key": {"value": 123, "another": [1, 2, {"more": "data"}]},
        }
        self.assertEqual(
            pretty_format(message),
            """\
8c668cde-235b-4872-af4e-caea524bd1c0 -> /1
2015-09-25T15:09:14Z
  key: {'another': [1, 2, {'more': 'data'}],
     |  'value': 123}
""",
        )

    def test_microsecond(self):
        """
        Microsecond timestamps are rendered in the output.
        """
        message = {
            "timestamp": 1443193754.123455,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
        }
        self.assertEqual(
            pretty_format(message),
            """\
8c668cde-235b-4872-af4e-caea524bd1c0 -> /1
2015-09-25T15:09:14.123455Z
""",
        )

    def test_compact(self):
        """
        The compact mode does everything on a single line, including
        dictionaries and multi-line messages.
        """
        message = {
            "timestamp": 1443193754,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
            "key": OrderedDict([("value", 123), ("another", [1, 2, {"more": "data"}])]),
            "multiline": "hello\n\tthere!\nabc",
        }
        self.assertEqual(
            compact_format(message),
            r'8c668cde-235b-4872-af4e-caea524bd1c0/1 2015-09-25T15:09:14Z key={"value":123,"another":[1,2,{"more":"data"}]} multiline="hello\n\tthere!\nabc"',
        )

    def test_local(self):
        """
        Timestamps can be generated in local timezone.
        """
        message = {
            "timestamp": 1443193754,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
        }
        expected = datetime.fromtimestamp(1443193754).isoformat(sep="T")
        self.assertIn(expected, pretty_format(message, True))
        self.assertIn(expected, compact_format(message, True))


class CommandLineTests(TestCase):
    """
    Tests for the command-line tool.
    """

    def test_help(self):
        """
        C{--help} prints out the help text and exits.
        """
        result = check_output(["eliot-prettyprint", "--help"])
        self.assertIn(b"Convert Eliot messages into more readable", result)

    def write_and_read(self, lines, extra_args=()):
        """
        Write the given lines to the command-line on stdin, return stdout.

        @param lines: Sequences of lines to write, as bytes, and lacking
            new lines.
        @return: Unicode-decoded result of subprocess stdout.
        """
        process = Popen(
            [b"eliot-prettyprint"] + list(extra_args), stdin=PIPE, stdout=PIPE
        )
        process.stdin.write(b"".join(line + b"\n" for line in lines))
        process.stdin.close()
        result = process.stdout.read().decode("utf-8")
        process.stdout.close()
        return result

    def test_output(self):
        """
        Lacking command-line arguments the process reads JSON lines from stdin
        and writes out a pretty-printed version.
        """
        messages = [SIMPLE_MESSAGE, UNTYPED_MESSAGE, SIMPLE_MESSAGE]
        stdout = self.write_and_read(map(dumps, messages))
        self.assertEqual(
            stdout, "".join(pretty_format(message) + "\n" for message in messages)
        )

    def test_compact_output(self):
        """
        In compact mode, the process reads JSON lines from stdin and writes out
        a pretty-printed compact version.
        """
        messages = [SIMPLE_MESSAGE, UNTYPED_MESSAGE, SIMPLE_MESSAGE]
        stdout = self.write_and_read(map(dumps, messages), [b"--compact"])
        self.assertEqual(
            stdout, "".join(compact_format(message) + "\n" for message in messages)
        )

    def test_local_timezone(self):
        """
        Local timezones are used if --local-timezone is given.
        """
        message = {
            "timestamp": 1443193754,
            "task_uuid": "8c668cde-235b-4872-af4e-caea524bd1c0",
            "task_level": [1],
        }
        expected = datetime.fromtimestamp(1443193754).isoformat(sep="T")
        stdout = self.write_and_read(
            [dumps(message)], [b"--compact", b"--local-timezone"]
        )
        self.assertIn(expected, stdout)
        stdout = self.write_and_read(
            [dumps(message)], [b"--compact", b"--local-timezone"]
        )
        self.assertIn(expected, stdout)

    def test_not_json_message(self):
        """
        Non-JSON lines are not formatted.
        """
        not_json = b"NOT JSON!!"
        lines = [dumps(SIMPLE_MESSAGE), not_json, dumps(UNTYPED_MESSAGE)]
        stdout = self.write_and_read(lines)
        self.assertEqual(
            stdout,
            "{}\nNot JSON: {}\n\n{}\n".format(
                pretty_format(SIMPLE_MESSAGE),
                str(not_json),
                pretty_format(UNTYPED_MESSAGE),
            ),
        )

    def test_missing_required_field(self):
        """
        Non-Eliot JSON messages are not formatted.
        """
        base = pmap(SIMPLE_MESSAGE)
        messages = [dumps(dict(base.remove(field))) for field in REQUIRED_FIELDS] + [
            dumps(SIMPLE_MESSAGE)
        ]
        stdout = self.write_and_read(messages)
        self.assertEqual(
            stdout,
            "{}{}\n".format(
                "".join(
                    "Not an Eliot message: {}\n\n".format(msg) for msg in messages[:-1]
                ),
                pretty_format(SIMPLE_MESSAGE),
            ),
        )



"""
Tests for L{eliot._action}.
"""

from __future__ import unicode_literals

import pickle
import time
from unittest import TestCase, skipIf
from unittest.mock import patch
from threading import Thread

import six

if six.PY3:
    unicode = six.text_type

from hypothesis import assume, given, settings, HealthCheck
from hypothesis.strategies import integers, lists, just, text

from pyrsistent import pvector, v

import testtools
from testtools.matchers import MatchesStructure

from .._action import (
    Action,
    current_action,
    startTask,
    start_action,
    ACTION_STATUS_FIELD,
    ACTION_TYPE_FIELD,
    FAILED_STATUS,
    STARTED_STATUS,
    SUCCEEDED_STATUS,
    DuplicateChild,
    InvalidStartMessage,
    InvalidStatus,
    TaskLevel,
    WrittenAction,
    WrongActionType,
    WrongTask,
    WrongTaskLevel,
    TooManyCalls,
    log_call,
)
from .._message import (
    EXCEPTION_FIELD,
    REASON_FIELD,
    TASK_LEVEL_FIELD,
    TASK_UUID_FIELD,
    MESSAGE_TYPE_FIELD,
    Message,
)
from .._output import MemoryLogger
from .._validation import ActionType, Field, _ActionSerializers
from ..testing import assertContainsFields, capture_logging
from ..parse import Parser
from .. import (
    add_destination,
    remove_destination,
    register_exception_extractor,
    preserve_context,
)

from .strategies import (
    message_dicts,
    start_action_message_dicts,
    start_action_messages,
    task_level_indexes,
    task_level_lists,
    written_actions,
    written_messages,
    reparent_action,
    sibling_task_level,
    union,
    written_from_pmap,
)


class ActionTests(TestCase):
    """
    Tests for L{Action}.
    """

    def test_start(self):
        """
        L{Action._start} logs an C{action_status="started"} message.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        action._start({"key": "value"})
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "unique",
                "task_level": [1],
                "action_type": "sys:thename",
                "action_status": "started",
                "key": "value",
            },
        )

    def test_task_uuid(self):
        """
        L{Action.task_uuid} return the task's UUID.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        self.assertEqual(action.task_uuid, "unique")

    def test_startMessageSerialization(self):
        """
        The start message logged by L{Action._start} is created with the
        appropriate start message L{eliot._validation._MessageSerializer}.
        """
        serializers = ActionType(
            "sys:thename", [Field("key", lambda x: x, "")], [], ""
        )._serializers

        class Logger(list):
            def write(self, msg, serializer):
                self.append(serializer)

        logger = Logger()
        action = Action(
            logger, "unique", TaskLevel(level=[]), "sys:thename", serializers
        )
        action._start({"key": "value"})
        self.assertIs(logger[0], serializers.start)

    def test_child(self):
        """
        L{Action.child} returns a new L{Action} with the given logger, system
        and name, and a task_uuid taken from the parent L{Action}.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        logger2 = MemoryLogger()
        child = action.child(logger2, "newsystem:newname")
        self.assertEqual(
            [child._logger, child._identification, child._task_level],
            [
                logger2,
                {"task_uuid": "unique", "action_type": "newsystem:newname"},
                TaskLevel(level=[1]),
            ],
        )

    def test_childLevel(self):
        """
        Each call to L{Action.child} increments the new sub-level set on the
        child.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        child1 = action.child(logger, "newsystem:newname")
        child2 = action.child(logger, "newsystem:newname")
        child1_1 = child1.child(logger, "newsystem:other")
        self.assertEqual(child1._task_level, TaskLevel(level=[1]))
        self.assertEqual(child2._task_level, TaskLevel(level=[2]))
        self.assertEqual(child1_1._task_level, TaskLevel(level=[1, 1]))

    def test_childSerializers(self):
        """
        L{Action.child} returns a new L{Action} with the serializers passed to
        it, rather than the parent's.
        """
        logger = MemoryLogger()
        serializers = object()
        action = Action(
            logger, "unique", TaskLevel(level=[]), "sys:thename", serializers
        )
        childSerializers = object()
        child = action.child(logger, "newsystem:newname", childSerializers)
        self.assertIs(child._serializers, childSerializers)

    def test_run(self):
        """
        L{Action.run} runs the given function with given arguments, returning
        its result.
        """
        action = Action(None, "", TaskLevel(level=[]), "")

        def f(*args, **kwargs):
            return args, kwargs

        result = action.run(f, 1, 2, x=3)
        self.assertEqual(result, ((1, 2), {"x": 3}))

    def test_runContext(self):
        """
        L{Action.run} runs the given function with the action set as the
        current action.
        """
        result = []
        action = Action(None, "", TaskLevel(level=[]), "")
        action.run(lambda: result.append(current_action()))
        self.assertEqual(result, [action])

    def test_per_thread_context(self):
        """Different threads have different contexts."""
        in_thread = []

        def run_in_thread():
            action = Action(None, "", TaskLevel(level=[]), "")
            with action.context():
                time.sleep(0.5)
                in_thread.append(current_action())

        thread = Thread(target=run_in_thread)
        thread.start()
        time.sleep(0.2)
        self.assertEqual(current_action(), None)
        thread.join()
        self.assertIsInstance(in_thread[0], Action)

    def test_runContextUnsetOnReturn(self):
        """
        L{Action.run} unsets the action once the given function returns.
        """
        action = Action(None, "", TaskLevel(level=[]), "")
        action.run(lambda: None)
        self.assertIs(current_action(), None)

    def test_runContextUnsetOnRaise(self):
        """
        L{Action.run} unsets the action once the given function raises an
        exception.
        """
        action = Action(None, "", TaskLevel(level=[]), "")
        self.assertRaises(ZeroDivisionError, action.run, lambda: 1 / 0)
        self.assertIs(current_action(), None)

    def test_withSetsContext(self):
        """
        L{Action.__enter__} sets the action as the current action.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        with action:
            self.assertIs(current_action(), action)

    def test_withUnsetOnReturn(self):
        """
        L{Action.__exit__} unsets the action on successful block finish.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        with action:
            pass
        self.assertIs(current_action(), None)

    def test_withUnsetOnRaise(self):
        """
        L{Action.__exit__} unsets the action if the block raises an exception.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        try:
            with action:
                1 / 0
        except ZeroDivisionError:
            pass
        else:
            self.fail("no exception")
        self.assertIs(current_action(), None)

    def test_withContextSetsContext(self):
        """
        L{Action.context().__enter__} sets the action as the current action.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        with action.context():
            self.assertIs(current_action(), action)

    def test_withContextReturnsaction(self):
        """
        L{Action.context().__enter__} returns the action.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        with action.context() as action2:
            self.assertIs(action, action2)

    def test_withContextUnsetOnReturn(self):
        """
        L{Action.context().__exit__} unsets the action on successful block
        finish.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        with action.context():
            pass
        self.assertIs(current_action(), None)

    def test_withContextNoLogging(self):
        """
        L{Action.context().__exit__} does not log any messages.
        """
        logger = MemoryLogger()
        action = Action(logger, "", TaskLevel(level=[]), "")
        with action.context():
            pass
        self.assertFalse(logger.messages)

    def test_withContextUnsetOnRaise(self):
        """
        L{Action.conext().__exit__} unsets the action if the block raises an
        exception.
        """
        action = Action(MemoryLogger(), "", TaskLevel(level=[]), "")
        try:
            with action.context():
                1 / 0
        except ZeroDivisionError:
            pass
        else:
            self.fail("no exception")
        self.assertIs(current_action(), None)

    def test_finish(self):
        """
        L{Action.finish} with no exception logs an C{action_status="succeeded"}
        message.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        action.finish()
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "unique",
                "task_level": [1],
                "action_type": "sys:thename",
                "action_status": "succeeded",
            },
        )

    def test_successfulFinishSerializer(self):
        """
        L{Action.finish} with no exception passes the success
        L{eliot._validation._MessageSerializer} to the message it creates.
        """
        serializers = ActionType(
            "sys:thename", [], [Field("key", lambda x: x, "")], ""
        )._serializers

        class Logger(list):
            def write(self, msg, serializer):
                self.append(serializer)

        logger = Logger()
        action = Action(
            logger, "unique", TaskLevel(level=[]), "sys:thename", serializers
        )
        action.finish()
        self.assertIs(logger[0], serializers.success)

    def test_failureFinishSerializer(self):
        """
        L{Action.finish} with an exception passes the failure
        L{eliot._validation._MessageSerializer} to the message it creates.
        """
        serializers = ActionType(
            "sys:thename", [], [Field("key", lambda x: x, "")], ""
        )._serializers

        class Logger(list):
            def write(self, msg, serializer):
                self.append(serializer)

        logger = Logger()
        action = Action(
            logger, "unique", TaskLevel(level=[]), "sys:thename", serializers
        )
        action.finish(Exception())
        self.assertIs(logger[0], serializers.failure)

    def test_startFieldsNotInFinish(self):
        """
        L{Action.finish} logs a message without the fields from
        L{Action._start}.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        action._start({"key": "value"})
        action.finish()
        self.assertNotIn("key", logger.messages[1])

    def test_finishWithBadException(self):
        """
        L{Action.finish} still logs a message if the given exception raises
        another exception when called with C{unicode()}.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")

        class BadException(Exception):
            def __str__(self):
                raise TypeError()

        action.finish(BadException())
        self.assertEqual(
            logger.messages[0]["reason"], "eliot: unknown, unicode() raised exception"
        )

    def test_withLogsSuccessfulFinishMessage(self):
        """
        L{Action.__exit__} logs an action finish message on a successful block
        finish.
        """
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action:
            pass
        # Start message is only created if we use the action()/task() utility
        # functions, the intended public APIs.
        self.assertEqual(len(logger.messages), 1)
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "uuid",
                "task_level": [1, 1],
                "action_type": "sys:me",
                "action_status": "succeeded",
            },
        )

    def test_withLogsExceptionMessage(self):
        """
        L{Action.__exit__} logs an action finish message on an exception
        raised from the block.
        """
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        exception = RuntimeError("because")

        try:
            with action:
                raise exception
        except RuntimeError:
            pass
        else:
            self.fail("no exception")

        self.assertEqual(len(logger.messages), 1)
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "uuid",
                "task_level": [1, 1],
                "action_type": "sys:me",
                "action_status": "failed",
                "reason": "because",
                "exception": "%s.RuntimeError" % (RuntimeError.__module__,),
            },
        )

    def test_withReturnValue(self):
        """
        L{Action.__enter__} returns the action itself.
        """
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action as act:
            self.assertIs(action, act)

    def test_addSuccessFields(self):
        """
        On a successful finish, L{Action.__exit__} adds fields from
        L{Action.addSuccessFields} to the result message.
        """
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action as act:
            act.addSuccessFields(x=1, y=2)
            act.addSuccessFields(z=3)
        assertContainsFields(self, logger.messages[0], {"x": 1, "y": 2, "z": 3})

    def test_nextTaskLevel(self):
        """
        Each call to L{Action._nextTaskLevel()} increments a counter.
        """
        action = Action(MemoryLogger(), "uuid", TaskLevel(level=[1]), "sys:me")
        self.assertEqual(
            [action._nextTaskLevel() for i in range(5)],
            [
                TaskLevel(level=level)
                for level in ([1, 1], [1, 2], [1, 3], [1, 4], [1, 5])
            ],
        )

    def test_multipleFinishCalls(self):
        """
        If L{Action.finish} is called, subsequent calls to L{Action.finish}
        have no effect.
        """
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action as act:
            act.finish()
            act.finish(Exception())
            act.finish()
        # Only initial finish message is logged:
        self.assertEqual(len(logger.messages), 1)


class StartActionAndTaskTests(TestCase):
    """
    Tests for L{start_action} and L{startTask}.
    """

    def test_startTaskNewAction(self):
        """
        L{startTask} creates a new top-level L{Action}.
        """
        logger = MemoryLogger()
        action = startTask(logger, "sys:do")
        self.assertIsInstance(action, Action)
        self.assertEqual(action._task_level, TaskLevel(level=[]))

    def test_start_task_default_action_type(self):
        """
        L{start_task} sets a default C{action_type} if none is set.
        """
        logger = MemoryLogger()
        startTask(logger)
        assertContainsFields(self, logger.messages[0], {"action_type": ""})

    def test_startTaskSerializers(self):
        """
        If serializers are passed to L{startTask} they are attached to the
        resulting L{Action}.
        """
        logger = MemoryLogger()
        serializers = _ActionSerializers(start=None, success=None, failure=None)
        action = startTask(logger, "sys:do", serializers)
        self.assertIs(action._serializers, serializers)

    def test_startActionSerializers(self):
        """
        If serializers are passed to L{start_action} they are attached to the
        resulting L{Action}.
        """
        logger = MemoryLogger()
        serializers = _ActionSerializers(start=None, success=None, failure=None)
        action = start_action(logger, "sys:do", serializers)
        self.assertIs(action._serializers, serializers)

    def test_startTaskNewUUID(self):
        """
        L{startTask} creates an L{Action} with its own C{task_uuid}.
        """
        logger = MemoryLogger()
        action = startTask(logger, "sys:do")
        action2 = startTask(logger, "sys:do")
        self.assertNotEqual(
            action._identification["task_uuid"], action2._identification["task_uuid"]
        )

    def test_startTaskLogsStart(self):
        """
        L{startTask} logs a start message for the newly created L{Action}.
        """
        logger = MemoryLogger()
        action = startTask(logger, "sys:do", key="value")
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": action._identification["task_uuid"],
                "task_level": [1],
                "action_type": "sys:do",
                "action_status": "started",
                "key": "value",
            },
        )

    def test_start_action_default_action_type(self):
        """
        L{start_action} sets a default C{action_type} if none is set.
        """
        logger = MemoryLogger()
        start_action(logger)
        assertContainsFields(self, logger.messages[0], {"action_type": ""})

    def test_startActionNoParent(self):
        """
        L{start_action} when C{current_action()} is C{None} creates a top-level
        L{Action}.
        """
        logger = MemoryLogger()
        action = start_action(logger, "sys:do")
        self.assertIsInstance(action, Action)
        self.assertEqual(action._task_level, TaskLevel(level=[]))

    def test_startActionNoParentLogStart(self):
        """
        L{start_action} when C{current_action()} is C{None} logs a start
        message.
        """
        logger = MemoryLogger()
        action = start_action(logger, "sys:do", key="value")
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": action._identification["task_uuid"],
                "task_level": [1],
                "action_type": "sys:do",
                "action_status": "started",
                "key": "value",
            },
        )

    def test_startActionWithParent(self):
        """
        L{start_action} uses the C{current_action()} as parent for a new
        L{Action}.
        """
        logger = MemoryLogger()
        parent = Action(logger, "uuid", TaskLevel(level=[2]), "other:thing")
        with parent:
            action = start_action(logger, "sys:do")
            self.assertIsInstance(action, Action)
            self.assertEqual(action._identification["task_uuid"], "uuid")
            self.assertEqual(action._task_level, TaskLevel(level=[2, 1]))

    def test_startActionWithParentLogStart(self):
        """
        L{start_action} when C{current_action()} is an L{Action} logs a start
        message.
        """
        logger = MemoryLogger()
        parent = Action(logger, "uuid", TaskLevel(level=[]), "other:thing")
        with parent:
            start_action(logger, "sys:do", key="value")
            assertContainsFields(
                self,
                logger.messages[0],
                {
                    "task_uuid": "uuid",
                    "task_level": [1, 1],
                    "action_type": "sys:do",
                    "action_status": "started",
                    "key": "value",
                },
            )

    def test_startTaskNoLogger(self):
        """
        When no logger is given L{startTask} logs to the default ``Logger``.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        action = startTask(action_type="sys:do", key="value")
        assertContainsFields(
            self,
            messages[0],
            {
                "task_uuid": action._identification["task_uuid"],
                "task_level": [1],
                "action_type": "sys:do",
                "action_status": "started",
                "key": "value",
            },
        )

    def test_startActionNoLogger(self):
        """
        When no logger is given L{start_action} logs to the default ``Logger``.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        action = start_action(action_type="sys:do", key="value")
        assertContainsFields(
            self,
            messages[0],
            {
                "task_uuid": action._identification["task_uuid"],
                "task_level": [1],
                "action_type": "sys:do",
                "action_status": "started",
                "key": "value",
            },
        )


class PEP8Tests(TestCase):
    """
    Tests for PEP 8 method compatibility.
    """

    def test_add_success_fields(self):
        """
        L{Action.addSuccessFields} is the same as L{Action.add_success_fields}.
        """
        self.assertEqual(Action.addSuccessFields, Action.add_success_fields)

    def test_serialize_task_id(self):
        """
        L{Action.serialize_task_id} is the same as L{Action.serializeTaskId}.
        """
        self.assertEqual(Action.serialize_task_id, Action.serializeTaskId)

    def test_continue_task(self):
        """
        L{Action.continue_task} is the same as L{Action.continueTask}.
        """
        self.assertEqual(Action.continue_task, Action.continueTask)


class SerializationTests(TestCase):
    """
    Tests for L{Action} serialization and deserialization.
    """

    def test_serializeTaskId(self):
        """
        L{Action.serialize_task_id} result is composed of the task UUID and an
        incremented task level.
        """
        action = Action(None, "uniq123", TaskLevel(level=[1, 2]), "mytype")
        self.assertEqual(
            [
                action._nextTaskLevel(),
                action.serialize_task_id(),
                action._nextTaskLevel(),
            ],
            [TaskLevel(level=[1, 2, 1]), b"uniq123@/1/2/2", TaskLevel(level=[1, 2, 3])],
        )

    def test_continueTaskReturnsAction(self):
        """
        L{Action.continue_task} returns an L{Action} whose C{task_level} and
        C{task_uuid} are derived from those in the given serialized task
        identifier.
        """
        originalAction = Action(None, "uniq456", TaskLevel(level=[3, 4]), "mytype")
        taskId = originalAction.serializeTaskId()

        newAction = Action.continue_task(MemoryLogger(), taskId)
        self.assertEqual(
            [newAction.__class__, newAction._identification, newAction._task_level],
            [
                Action,
                {"task_uuid": "uniq456", "action_type": "eliot:remote_task"},
                TaskLevel(level=[3, 4, 1]),
            ],
        )

    def test_continueTaskUnicode(self):
        """
        L{Action.continue_task} can take a Unicode task identifier.
        """
        original_action = Action(None, "uniq790", TaskLevel(level=[3, 4]), "mytype")
        task_id = unicode(original_action.serialize_task_id(), "utf-8")

        new_action = Action.continue_task(MemoryLogger(), task_id)
        self.assertEqual(new_action._identification["task_uuid"], "uniq790")

    def test_continueTaskStartsAction(self):
        """
        L{Action.continue_task} starts the L{Action} it creates.
        """
        originalAction = Action(None, "uniq456", TaskLevel(level=[3, 4]), "mytype")
        taskId = originalAction.serializeTaskId()
        logger = MemoryLogger()

        Action.continue_task(logger, taskId)
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "uniq456",
                "task_level": [3, 4, 1, 1],
                "action_type": "eliot:remote_task",
                "action_status": "started",
            },
        )

    def test_continueTaskNoLogger(self):
        """
        L{Action.continue_task} can be called without a logger.
        """
        originalAction = Action(None, "uniq456", TaskLevel(level=[3, 4]), "mytype")
        taskId = originalAction.serializeTaskId()

        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        Action.continue_task(task_id=taskId)
        assertContainsFields(
            self,
            messages[-1],
            {
                "task_uuid": "uniq456",
                "task_level": [3, 4, 1, 1],
                "action_type": "eliot:remote_task",
                "action_status": "started",
            },
        )

    def test_continueTaskRequiredTaskId(self):
        """
        L{Action.continue_task} requires a C{task_id} to be passed in.
        """
        self.assertRaises(RuntimeError, Action.continue_task)


class TaskLevelTests(TestCase):
    """
    Tests for L{TaskLevel}.
    """

    def assert_fully_less_than(self, x, y):
        """
        Assert that x < y according to all the comparison operators.
        """
        self.assertTrue(
            all(
                [
                    # lt
                    x < y,
                    not y < x,
                    # le
                    x <= y,
                    not y <= x,
                    # gt
                    y > x,
                    not x > y,
                    # ge
                    y >= x,
                    not x >= y,
                    # eq
                    not x == y,
                    not y == x,
                    # ne
                    x != y,
                    y != x,
                ]
            )
        )

    def test_equality(self):
        """
        L{TaskChild} correctly implements equality and hashing.
        """
        a = TaskLevel(level=[1, 2])
        a2 = TaskLevel(level=[1, 2])
        b = TaskLevel(level=[2, 999])
        self.assertTrue(
            all(
                [
                    a == a2,
                    a2 == a,
                    a != b,
                    b != a,
                    not b == a,
                    not a == b,
                    not a == 1,
                    a != 1,
                    hash(a) == hash(a2),
                    hash(b) != hash(a),
                ]
            )
        )

    def test_as_list(self):
        """
        L{TaskChild.as_list} returns the level.
        """
        self.assertEqual(TaskLevel(level=[1, 2, 3]).as_list(), [1, 2, 3])

    @given(lists(task_level_indexes))
    def test_parent_of_child(self, base_task_level):
        """
        L{TaskLevel.child} returns the first child of the task.
        """
        base_task = TaskLevel(level=base_task_level)
        child_task = base_task.child()
        self.assertEqual(base_task, child_task.parent())

    @given(task_level_lists)
    def test_child_greater_than_parent(self, task_level):
        """
        L{TaskLevel.child} returns a child that is greater than its parent.
        """
        task = TaskLevel(level=task_level)
        self.assert_fully_less_than(task, task.child())

    @given(task_level_lists)
    def test_next_sibling_greater(self, task_level):
        """
        L{TaskLevel.next_sibling} returns a greater task level.
        """
        task = TaskLevel(level=task_level)
        self.assert_fully_less_than(task, task.next_sibling())

    @given(task_level_lists)
    def test_next_sibling(self, task_level):
        """
        L{TaskLevel.next_sibling} returns the next sibling of a task.
        """
        task = TaskLevel(level=task_level)
        sibling = task.next_sibling()
        self.assertEqual(
            sibling, TaskLevel(level=task_level[:-1] + [task_level[-1] + 1])
        )

    def test_parent_of_root(self):
        """
        L{TaskLevel.parent} of the root task level is C{None}.
        """
        self.assertIs(TaskLevel(level=[]).parent(), None)

    def test_toString(self):
        """
        L{TaskLevel.toString} serializes the object to a Unicode string.
        """
        root = TaskLevel(level=[])
        child2_1 = root.child().next_sibling().child()
        self.assertEqual([root.toString(), child2_1.toString()], ["/", "/2/1"])

    def test_fromString(self):
        """
        L{TaskLevel.fromString} deserializes the output of
        L{TaskLevel.toString}.
        """
        self.assertEqual(
            [TaskLevel.fromString("/"), TaskLevel.fromString("/2/1")],
            [TaskLevel(level=[]), TaskLevel(level=[2, 1])],
        )

    def test_from_string(self):
        """
        L{TaskLevel.from_string} is the same as as L{TaskLevel.fromString}.
        """
        self.assertEqual(TaskLevel.from_string, TaskLevel.fromString)

    def test_to_string(self):
        """
        L{TaskLevel.to_string} is the same as as L{TaskLevel.toString}.
        """
        self.assertEqual(TaskLevel.to_string, TaskLevel.toString)


class WrittenActionTests(testtools.TestCase):
    """
    Tests for L{WrittenAction}.
    """

    @given(start_action_messages)
    def test_from_single_start_message(self, message):
        """
        A L{WrittenAction} can be constructed from a single "start" message.
        Such an action inherits the C{action_type} of the start message, has no
        C{end_time}, and has a C{status} of C{STARTED_STATUS}.
        """
        action = WrittenAction.from_messages(message)
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                status=STARTED_STATUS,
                action_type=message.contents[ACTION_TYPE_FIELD],
                task_uuid=message.task_uuid,
                task_level=message.task_level.parent(),
                start_time=message.timestamp,
                children=pvector([]),
                end_time=None,
                reason=None,
                exception=None,
            ),
        )

    @given(start_action_messages, message_dicts, integers(min_value=2))
    def test_from_single_end_message(self, start_message, end_message_dict, n):
        """
        A L{WrittenAction} can be constructed from a single "end"
        message. Such an action inherits the C{action_type} and
        C{task_level} of the end message, has no C{start_time}, and has a
        C{status} matching that of the end message.
        """
        end_message = written_from_pmap(
            union(
                end_message_dict,
                {
                    ACTION_STATUS_FIELD: SUCCEEDED_STATUS,
                    ACTION_TYPE_FIELD: start_message.contents[ACTION_TYPE_FIELD],
                    TASK_UUID_FIELD: start_message.task_uuid,
                    TASK_LEVEL_FIELD: sibling_task_level(start_message, n),
                },
            )
        )
        action = WrittenAction.from_messages(end_message=end_message)
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                status=SUCCEEDED_STATUS,
                action_type=end_message.contents[ACTION_TYPE_FIELD],
                task_uuid=end_message.task_uuid,
                task_level=end_message.task_level.parent(),
                start_time=None,
                children=pvector([]),
                end_time=end_message.timestamp,
                reason=None,
                exception=None,
            ),
        )

    @given(message_dicts)
    def test_from_single_child_message(self, message_dict):
        """
        A L{WrittenAction} can be constructed from a single child
        message. Such an action inherits the C{task_level} of the message,
        has no C{start_time}, C{status}, C{task_type} or C{end_time}.
        """
        message = written_from_pmap(message_dict)
        action = WrittenAction.from_messages(children=[message])
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                status=None,
                action_type=None,
                task_uuid=message.task_uuid,
                task_level=message.task_level.parent(),
                start_time=None,
                children=pvector([message]),
                end_time=None,
                reason=None,
                exception=None,
            ),
        )

    @given(start_action_messages, message_dicts, integers(min_value=2))
    def test_different_task_uuid(self, start_message, end_message_dict, n):
        """
        By definition, an action is either a top-level task or takes place
        within such a task. If we try to assemble actions from messages with
        differing task UUIDs, we raise an error.
        """
        assume(start_message.task_uuid != end_message_dict["task_uuid"])
        action_type = start_message.as_dict()[ACTION_TYPE_FIELD]
        end_message = written_from_pmap(
            union(
                end_message_dict.set(ACTION_TYPE_FIELD, action_type),
                {
                    ACTION_STATUS_FIELD: SUCCEEDED_STATUS,
                    TASK_LEVEL_FIELD: sibling_task_level(start_message, n),
                },
            )
        )
        self.assertRaises(
            WrongTask,
            WrittenAction.from_messages,
            start_message,
            end_message=end_message,
        )

    @given(message_dicts)
    def test_invalid_start_message_missing_status(self, message_dict):
        """
        A start message must have an C{ACTION_STATUS_FIELD} with the value
        C{STARTED_STATUS}, otherwise it's not a start message. If we receive
        such a message, raise an error.

        This test handles the case where the status field is not present.
        """
        assume(ACTION_STATUS_FIELD not in message_dict)
        message = written_from_pmap(message_dict)
        self.assertRaises(InvalidStartMessage, WrittenAction.from_messages, message)

    @given(
        message_dict=start_action_message_dicts,
        status=(just(FAILED_STATUS) | just(SUCCEEDED_STATUS) | text()),
    )
    def test_invalid_start_message_wrong_status(self, message_dict, status):
        """
        A start message must have an C{ACTION_STATUS_FIELD} with the value
        C{STARTED_STATUS}, otherwise it's not a start message. If we receive
        such a message, raise an error.

        This test handles the case where the status field is present, but is
        not C{STARTED_STATUS}.
        """
        message = written_from_pmap(message_dict.update({ACTION_STATUS_FIELD: status}))
        self.assertRaises(InvalidStartMessage, WrittenAction.from_messages, message)

    @given(start_action_message_dicts, integers(min_value=2))
    def test_invalid_task_level_in_start_message(self, start_message_dict, i):
        """
        All messages in an action have a task level. The first message in an
        action must have a task level ending in C{1}, indicating that it's the
        first message.

        If we try to start an action with a message that has a task level that
        does not end in C{1}, raise an error.
        """
        new_level = start_message_dict[TASK_LEVEL_FIELD].append(i)
        message_dict = start_message_dict.set(TASK_LEVEL_FIELD, new_level)
        message = written_from_pmap(message_dict)
        self.assertRaises(InvalidStartMessage, WrittenAction.from_messages, message)

    @given(start_action_messages, message_dicts, text(), integers(min_value=1))
    def test_action_type_mismatch(self, start_message, end_message_dict, end_type, n):
        """
        The end message of an action must have the same C{ACTION_TYPE_FIELD} as
        the start message of an action. If we try to end an action with a
        message that has a different type, we raise an error.
        """
        assume(end_type != start_message.contents[ACTION_TYPE_FIELD])
        end_message = written_from_pmap(
            union(
                end_message_dict,
                {
                    ACTION_STATUS_FIELD: SUCCEEDED_STATUS,
                    ACTION_TYPE_FIELD: end_type,
                    TASK_UUID_FIELD: start_message.task_uuid,
                    TASK_LEVEL_FIELD: sibling_task_level(start_message, n),
                },
            )
        )
        self.assertRaises(
            WrongActionType,
            WrittenAction.from_messages,
            start_message,
            end_message=end_message,
        )

    @given(start_action_messages, message_dicts, integers(min_value=2))
    def test_successful_end(self, start_message, end_message_dict, n):
        """
        A L{WrittenAction} can be constructed with just a start message and an
        end message: in this case, an end message that indicates the action was
        successful.

        Such an action inherits the C{end_time} from the end message, and has
        a C{status} of C{SUCCEEDED_STATUS}.
        """
        end_message = written_from_pmap(
            union(
                end_message_dict,
                {
                    ACTION_STATUS_FIELD: SUCCEEDED_STATUS,
                    ACTION_TYPE_FIELD: start_message.contents[ACTION_TYPE_FIELD],
                    TASK_UUID_FIELD: start_message.task_uuid,
                    TASK_LEVEL_FIELD: sibling_task_level(start_message, n),
                },
            )
        )
        action = WrittenAction.from_messages(start_message, end_message=end_message)
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                action_type=start_message.contents[ACTION_TYPE_FIELD],
                status=SUCCEEDED_STATUS,
                task_uuid=start_message.task_uuid,
                task_level=start_message.task_level.parent(),
                start_time=start_message.timestamp,
                children=pvector([]),
                end_time=end_message.timestamp,
                reason=None,
                exception=None,
            ),
        )

    @given(start_action_messages, message_dicts, text(), text(), integers(min_value=2))
    def test_failed_end(self, start_message, end_message_dict, exception, reason, n):
        """
        A L{WrittenAction} can be constructed with just a start message and an
        end message: in this case, an end message that indicates that the
        action failed.

        Such an action inherits the C{end_time} from the end message, has a
        C{status} of C{FAILED_STATUS}, and an C{exception} and C{reason} that
        match the raised exception.
        """
        end_message = written_from_pmap(
            union(
                end_message_dict,
                {
                    ACTION_STATUS_FIELD: FAILED_STATUS,
                    ACTION_TYPE_FIELD: start_message.contents[ACTION_TYPE_FIELD],
                    TASK_UUID_FIELD: start_message.task_uuid,
                    TASK_LEVEL_FIELD: sibling_task_level(start_message, n),
                    EXCEPTION_FIELD: exception,
                    REASON_FIELD: reason,
                },
            )
        )
        action = WrittenAction.from_messages(start_message, end_message=end_message)
        self.assertThat(
            action,
            MatchesStructure.byEquality(
                action_type=start_message.contents[ACTION_TYPE_FIELD],
                status=FAILED_STATUS,
                task_uuid=start_message.task_uuid,
                task_level=start_message.task_level.parent(),
                start_time=start_message.timestamp,
                children=pvector([]),
                end_time=end_message.timestamp,
                reason=reason,
                exception=exception,
            ),
        )

    @given(start_action_messages, message_dicts, integers(min_value=2))
    def test_end_has_no_status(self, start_message, end_message_dict, n):
        """
        If we try to end a L{WrittenAction} with a message that lacks an
        C{ACTION_STATUS_FIELD}, we raise an error, because it's not a valid
        end message.
        """
        assume(ACTION_STATUS_FIELD not in end_message_dict)
        end_message = written_from_pmap(
            union(
                end_message_dict,
                {
                    ACTION_TYPE_FIELD: start_message.contents[ACTION_TYPE_FIELD],
                    TASK_UUID_FIELD: start_message.task_uuid,
                    TASK_LEVEL_FIELD: sibling_task_level(start_message, n),
                },
            )
        )
        self.assertRaises(
            InvalidStatus,
            WrittenAction.from_messages,
            start_message,
            end_message=end_message,
        )

    # This test is slow, and when run under coverage on pypy on Travis won't
    # make the default of 5 examples. 1 is enough.
    @given(start_action_messages, lists(written_messages | written_actions))
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_children(self, start_message, child_messages):
        """
        We can construct a L{WrittenAction} with child messages. These messages
        can be either L{WrittenAction}s or L{WrittenMessage}s. They are
        available in the C{children} field.
        """
        messages = [
            reparent_action(
                start_message.task_uuid,
                TaskLevel(level=sibling_task_level(start_message, i)),
                message,
            )
            for (i, message) in enumerate(child_messages, 2)
        ]
        action = WrittenAction.from_messages(start_message, messages)

        def task_level(m):
            return m.task_level

        self.assertEqual(sorted(messages, key=task_level), action.children)

    @given(start_action_messages, message_dicts)
    def test_wrong_task_uuid(self, start_message, child_message):
        """
        All child messages of an action must have the same C{task_uuid} as the
        action.
        """
        assume(child_message[TASK_UUID_FIELD] != start_message.task_uuid)
        message = written_from_pmap(child_message)
        self.assertRaises(
            WrongTask, WrittenAction.from_messages, start_message, v(message)
        )

    @given(start_action_messages, message_dicts)
    def test_wrong_task_level(self, start_message, child_message):
        """
        All child messages of an action must have a task level that is a direct
        child of the action's task level.
        """
        assume(
            not start_message.task_level.is_sibling_of(
                TaskLevel(level=child_message[TASK_LEVEL_FIELD])
            )
        )
        message = written_from_pmap(
            child_message.update({TASK_UUID_FIELD: start_message.task_uuid})
        )
        self.assertRaises(
            WrongTaskLevel, WrittenAction.from_messages, start_message, v(message)
        )

    @given(start_action_messages, message_dicts, message_dicts, integers(min_value=2))
    def test_duplicate_task_level(self, start_message, child1, child2, index):
        """
        If we try to add a child to an action that has a task level that's the
        same as the task level of an existing child, we raise an error.
        """
        parent_level = start_message.task_level.parent().level
        messages = [
            written_from_pmap(
                union(
                    child_message,
                    {
                        TASK_UUID_FIELD: start_message.task_uuid,
                        TASK_LEVEL_FIELD: parent_level.append(index),
                    },
                )
            )
            for child_message in [child1, child2]
        ]
        assume(messages[0] != messages[1])
        self.assertRaises(
            DuplicateChild, WrittenAction.from_messages, start_message, messages
        )


def make_error_extraction_tests(get_messages):
    """
    Create a test case class for testing extraction of fields from exceptions.

    @param get_messages: Callable that takes an exception instance, returns
        all message dictionaries generated by logging it.

    @return: ``TestCase`` subclass.
    """

    class ErrorFieldExtraction(TestCase):
        """
        Tests for extracting fields from exceptions in failed actions.
        """

        def test_matching_class(self):
            """
            If an exception fails an action and the exact type has registered
            extractor, extract errors using it.
            """

            class MyException(Exception):
                pass

            register_exception_extractor(MyException, lambda e: {"key": e.args[0]})
            exception = MyException("a value")
            [message] = get_messages(exception)
            assertContainsFields(self, message, {"key": "a value"})

        def test_subclass_falls_back_to_parent(self):
            """
            If an exception fails an action and the exact type has not been
            registered but the error is a subclass of a registered class,
            extract errors using it.
            """

            class MyException(Exception):
                pass

            class SubException(MyException):
                pass

            register_exception_extractor(MyException, lambda e: {"key": e.args[0]})
            [message] = get_messages(SubException("the value"))
            assertContainsFields(self, message, {"key": "the value"})

        def test_subclass_matches_first(self):
            """
            If both a superclass and base class have registered extractors, the
            more specific one is used.
            """

            class MyException(Exception):
                pass

            class SubException(MyException):
                pass

            class SubSubException(SubException):
                pass

            register_exception_extractor(MyException, lambda e: {"parent": e.args[0]})
            register_exception_extractor(SubException, lambda e: {"child": e.args[0]})
            [message] = get_messages(SubSubException("the value"))
            assertContainsFields(self, message, {"child": "the value"})

        def test_error_in_extracter(self):
            """
            If an error occurs in extraction, log the message as usual just
            without the extra fields, and an additional traceback.
            """

            class MyException(Exception):
                pass

            def extract(e):
                return e.nosuchattribute

            register_exception_extractor(MyException, extract)

            messages = get_failed_action_messages(MyException())
            assertContainsFields(
                self, messages[1], {"action_type": "sys:me", "action_status": "failed"}
            )
            assertContainsFields(self, messages[0], {"message_type": "eliot:traceback"})
            self.assertIn("nosuchattribute", str(messages[0]["reason"]))

        def test_environmenterror(self):
            """
            ``EnvironmentError`` has a registered extractor that extracts the
            errno.
            """
            [message] = get_messages(EnvironmentError(12, "oh noes"))
            assertContainsFields(self, message, {"errno": 12})

    return ErrorFieldExtraction


def get_failed_action_messages(exception):
    """
    Fail an action using the given exception.

    :return: Logged dictionaries from the exception failing an action.
    """
    action_type = ActionType("sys:me", [], [])
    logger = MemoryLogger()
    action = action_type.as_task(logger=logger)
    try:
        with action:
            raise exception
    except exception.__class__:
        pass
    logger.validate()
    return logger.messages[1:]


class FailedActionExtractionTests(
    make_error_extraction_tests(get_failed_action_messages)
):
    """
    Tests for error extraction in failed actions.
    """

    def test_regular_fields(self):
        """
        The normal failed action fields are still present when error
        extraction is used.
        """

        class MyException(Exception):
            pass

        register_exception_extractor(MyException, lambda e: {"key": e.args[0]})

        exception = MyException("because")
        messages = get_failed_action_messages(exception)
        assertContainsFields(
            self,
            messages[0],
            {
                "task_level": [2],
                "action_type": "sys:me",
                "action_status": "failed",
                "reason": "because",
                "exception": "eliot.tests.test_action.MyException",
            },
        )


class PreserveContextTests(TestCase):
    """
    Tests for L{preserve_context}.
    """

    def add(self, x, y):
        """
        Add two inputs.
        """
        Message.log(message_type="child")
        return x + y

    def test_no_context(self):
        """
        If C{preserve_context} is run outside an action context it just
        returns the same function.
        """
        wrapped = preserve_context(self.add)
        self.assertEqual(wrapped(2, 3), 5)

    def test_with_context_calls_underlying(self):
        """
        If run inside an Eliot context, the result of C{preserve_context} is
        the result of calling the underlying function.
        """
        with start_action(action_type="parent"):
            wrapped = preserve_context(self.add)
            self.assertEqual(wrapped(3, y=4), 7)

    @capture_logging(None)
    def test_with_context_preserves_context(self, logger):
        """
        If run inside an Eliot context, the result of C{preserve_context} runs
        the wrapped function within a C{eliot:task} which is a child of
        the original action.
        """
        with start_action(action_type="parent"):
            wrapped = preserve_context(lambda: self.add(3, 4))
        thread = Thread(target=wrapped)
        thread.start()
        thread.join()
        [tree] = Parser.parse_stream(logger.messages)
        root = tree.root()
        self.assertEqual(
            (
                root.action_type,
                root.children[0].action_type,
                root.children[0].children[0].contents[MESSAGE_TYPE_FIELD],
            ),
            ("parent", "eliot:remote_task", "child"),
        )

    def test_callable_only_once(self):
        """
        The result of C{preserve_context} can only be called once.
        """
        with start_action(action_type="parent"):
            wrapped = preserve_context(self.add)
        wrapped(1, 2)
        self.assertRaises(TooManyCalls, wrapped, 3, 4)


@log_call
def for_pickling():
    pass


class LogCallTests(TestCase):
    """Tests for log_call decorator."""

    def assert_logged(self, logger, action_type, expected_params, expected_result):
        """Assert that an action of given structure was logged."""
        if six.PY2:
            # On Python 2 we don't include the module or class:
            action_type = action_type.split(".")[-1]
        [tree] = Parser.parse_stream(logger.messages)
        root = tree.root()
        self.assertEqual(root.action_type, action_type)
        message = dict(root.start_message.contents)
        for field in [ACTION_STATUS_FIELD, ACTION_TYPE_FIELD]:
            message.pop(field)
        self.assertEqual(message, expected_params)
        self.assertEqual(root.end_message.contents["result"], expected_result)
        self.assertEqual(root.status, SUCCEEDED_STATUS)

    @capture_logging(None)
    def test_no_args_return(self, logger):
        """
        C{@log_call} with no arguments logs return result, arguments, and has
        action type based on the action name.
        """

        @log_call
        def myfunc(x, y):
            return 4

        myfunc(2, 3)
        self.assert_logged(logger, self.id() + ".<locals>.myfunc", {"x": 2, "y": 3}, 4)

    @capture_logging(None)
    def test_exception(self, logger):
        """C{@log_call} with an exception logs a failed action."""

        @log_call
        def myfunc(x, y):
            1 / 0

        with self.assertRaises(ZeroDivisionError):
            myfunc(2, 4)

        [tree] = Parser.parse_stream(logger.messages)
        root = tree.root()
        self.assertIn("ZeroDivisionError", root.end_message.contents["exception"])
        self.assertEqual(root.status, FAILED_STATUS)

    @capture_logging(None)
    def test_action_type(self, logger):
        """C{@log_call} can take an action type."""

        @log_call(action_type="myaction")
        def myfunc(x, y):
            return 4

        myfunc(2, 3)
        self.assert_logged(logger, "myaction", {"x": 2, "y": 3}, 4)

    @capture_logging(None)
    def test_default_argument_given(self, logger):
        """C{@log_call} logs default arguments that were passed in."""

        @log_call
        def myfunc(x, y=1):
            return 4

        myfunc(2, y=5)
        self.assert_logged(logger, self.id() + ".<locals>.myfunc", {"x": 2, "y": 5}, 4)

    @capture_logging(None)
    def test_default_argument_missing(self, logger):
        """C{@log_call} logs default arguments that weren't passed in."""

        @log_call
        def myfunc(x, y=1):
            return 6

        myfunc(2)
        self.assert_logged(logger, self.id() + ".<locals>.myfunc", {"x": 2, "y": 1}, 6)

    @capture_logging(None)
    def test_star_args_kwargs(self, logger):
        """C{@log_call} logs star args and kwargs."""

        @log_call
        def myfunc(x, *y, **z):
            return 6

        myfunc(2, 3, 4, a=1, b=2)
        self.assert_logged(
            logger,
            self.id() + ".<locals>.myfunc",
            {"x": 2, "y": (3, 4), "z": {"a": 1, "b": 2}},
            6,
        )

    @capture_logging(None)
    def test_whitelist_args(self, logger):
        """C{@log_call} only includes whitelisted arguments."""

        @log_call(include_args=("x", "z"))
        def myfunc(x, y, z):
            return 6

        myfunc(2, 3, 4)
        self.assert_logged(logger, self.id() + ".<locals>.myfunc", {"x": 2, "z": 4}, 6)

    @skipIf(six.PY2, "Didn't bother implementing safety check on Python 2")
    def test_wrong_whitelist_args(self):
        """If C{include_args} doesn't match function, raise an exception."""
        with self.assertRaises(ValueError):

            @log_call(include_args=["a", "x", "y"])
            def f(x, y):
                pass

    @capture_logging(None)
    def test_no_result(self, logger):
        """C{@log_call} can omit logging the result."""

        @log_call(include_result=False)
        def myfunc(x, y):
            return 6

        myfunc(2, 3)

        [tree] = Parser.parse_stream(logger.messages)
        root = tree.root()
        self.assertNotIn("result", root.end_message.contents)
        self.assertEqual(root.status, SUCCEEDED_STATUS)

    def test_pickleable(self):
        """Functions decorated with C{log_call} are pickleable.

        This is necessary for e.g. Dask usage.
        """
        self.assertIs(for_pickling, pickle.loads(pickle.dumps(for_pickling)))

    @capture_logging(None)
    def test_methods(self, logger):
        """self is not logged."""

        class C(object):
            @log_call
            def f(self, x):
                pass

        C().f(2)
        self.assert_logged(logger, self.id() + ".<locals>.C.f", {"x": 2}, None)


class IndividualMessageLogTests(TestCase):
    """Action.log() tests."""

    def test_log_creates_new_dictionary(self):
        """
        L{Action.log} creates a new dictionary on each call.

        This is important because we might mutate the dictionary in
        ``Logger.write``.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)

        with start_action(action_type="x") as action:
            action.log("mymessage", key=4)
            action.log(message_type="mymessage2", key=5)
        self.assertEqual(messages[1]["key"], 4)
        self.assertEqual(messages[2]["key"], 5)
        self.assertEqual(messages[1]["message_type"], "mymessage")
        self.assertEqual(messages[2]["message_type"], "mymessage2")

    @patch("time.time")
    def test_log_adds_timestamp(self, time_func):
        """
        L{Action.log} adds a C{"timestamp"} field to the dictionary written
        to the logger, with the current time in seconds since the epoch.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)

        time_func.return_value = timestamp = 1387299889.153187625
        with start_action(action_type="x") as action:
            action.log("mymessage", key=4)
        self.assertEqual(messages[1]["timestamp"], timestamp)

    def test_part_of_action(self):
        """
        L{Action.log} adds the identification fields from the given
        L{Action} to the dictionary written to the logger.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)

        action = Action(None, "unique", TaskLevel(level=[37, 4]), "sys:thename")
        action.log("me", key=2)
        written = messages[0]
        del written["timestamp"]
        self.assertEqual(
            written,
            {
                "task_uuid": "unique",
                "task_level": [37, 4, 1],
                "key": 2,
                "message_type": "me",
            },
        )



"""Integration with the standard library ``logging`` package."""

from logging import Handler

from ._action import log_message
from ._traceback import write_traceback


class EliotHandler(Handler):
    """A C{logging.Handler} that routes log messages to Eliot."""

    def emit(self, record):
        log_message(
            message_type="eliot:stdlib",
            log_level=record.levelname,
            logger=record.name,
            message=record.getMessage(),
        )
        if record.exc_info:
            write_traceback(exc_info=record.exc_info)


__all__ = ["EliotHandler"]



"""
Tests for L{eliot._generators}.
"""

from __future__ import unicode_literals, absolute_import

from pprint import pformat
from unittest import TestCase

from eliot import Message, start_action
from ..testing import capture_logging, assertHasAction

from .._generators import eliot_friendly_generator_function


def assert_expected_action_tree(
    testcase, logger, expected_action_type, expected_type_tree
):
    """
    Assert that a logger has a certain logged action with certain children.

    @see: L{assert_generator_logs_action_tree}
    """
    logged_action = assertHasAction(testcase, logger, expected_action_type, True)
    type_tree = logged_action.type_tree()
    testcase.assertEqual(
        {expected_action_type: expected_type_tree},
        type_tree,
        "Logger had messages:\n{}".format(pformat(logger.messages, indent=4)),
    )


def assert_generator_logs_action_tree(
    testcase, generator_function, logger, expected_action_type, expected_type_tree
):
    """
    Assert that exhausting a generator from the given function logs an action
    of the given type with children matching the given type tree.

    @param testcase: A test case instance to use to make assertions.
    @type testcase: L{unittest.TestCase}

    @param generator_function: A no-argument callable that returns a generator
        to be exhausted.

    @param logger: A logger to inspect for logged messages.
    @type logger: L{MemoryLogger}

    @param expected_action_type: An action type which should be logged by the
        generator.
    @type expected_action_type: L{unicode}

    @param expected_type_tree: The types of actions and messages which should
        be logged beneath the expected action.  The structure of this value
        matches the structure returned by L{LoggedAction.type_tree}.
    @type expected_type_tree: L{list}
    """
    list(eliot_friendly_generator_function(generator_function)())
    assert_expected_action_tree(
        testcase, logger, expected_action_type, expected_type_tree
    )


class EliotFriendlyGeneratorFunctionTests(TestCase):
    """
    Tests for L{eliot_friendly_generator_function}.
    """

    # Get our custom assertion failure messages *and* the standard ones.
    longMessage = True

    @capture_logging(None)
    def test_yield_none(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type="hello")
            yield
            Message.log(message_type="goodbye")

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            list(g())

        assert_expected_action_tree(
            self, logger, "the-action", ["hello", "yielded", "goodbye"]
        )

    @capture_logging(None)
    def test_yield_value(self, logger):
        expected = object()

        @eliot_friendly_generator_function
        def g():
            Message.log(message_type="hello")
            yield expected
            Message.log(message_type="goodbye")

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            self.assertEqual([expected], list(g()))

        assert_expected_action_tree(
            self, logger, "the-action", ["hello", "yielded", "goodbye"]
        )

    @capture_logging(None)
    def test_yield_inside_another_action(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type="a")
            with start_action(action_type="confounding-factor"):
                Message.log(message_type="b")
                yield None
                Message.log(message_type="c")
            Message.log(message_type="d")

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            list(g())

        assert_expected_action_tree(
            self,
            logger,
            "the-action",
            ["a", {"confounding-factor": ["b", "yielded", "c"]}, "d"],
        )

    @capture_logging(None)
    def test_yield_inside_nested_actions(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type="a")
            with start_action(action_type="confounding-factor"):
                Message.log(message_type="b")
                yield None
                with start_action(action_type="double-confounding-factor"):
                    yield None
                    Message.log(message_type="c")
                Message.log(message_type="d")
            Message.log(message_type="e")

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            list(g())

        assert_expected_action_tree(
            self,
            logger,
            "the-action",
            [
                "a",
                {
                    "confounding-factor": [
                        "b",
                        "yielded",
                        {"double-confounding-factor": ["yielded", "c"]},
                        "d",
                    ]
                },
                "e",
            ],
        )

    @capture_logging(None)
    def test_generator_and_non_generator(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type="a")
            yield
            with start_action(action_type="action-a"):
                Message.log(message_type="b")
                yield
                Message.log(message_type="c")

            Message.log(message_type="d")
            yield

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            generator = g()
            next(generator)
            Message.log(message_type="0")
            next(generator)
            Message.log(message_type="1")
            next(generator)
            Message.log(message_type="2")
            self.assertRaises(StopIteration, lambda: next(generator))

        assert_expected_action_tree(
            self,
            logger,
            "the-action",
            [
                "a",
                "yielded",
                "0",
                {"action-a": ["b", "yielded", "c"]},
                "1",
                "d",
                "yielded",
                "2",
            ],
        )

    @capture_logging(None)
    def test_concurrent_generators(self, logger):
        @eliot_friendly_generator_function
        def g(which):
            Message.log(message_type="{}-a".format(which))
            with start_action(action_type=which):
                Message.log(message_type="{}-b".format(which))
                yield
                Message.log(message_type="{}-c".format(which))
            Message.log(message_type="{}-d".format(which))

        g.debug = True  # output yielded messages

        gens = [g("1"), g("2")]
        with start_action(action_type="the-action"):
            while gens:
                for g in gens[:]:
                    try:
                        next(g)
                    except StopIteration:
                        gens.remove(g)

        assert_expected_action_tree(
            self,
            logger,
            "the-action",
            [
                "1-a",
                {"1": ["1-b", "yielded", "1-c"]},
                "2-a",
                {"2": ["2-b", "yielded", "2-c"]},
                "1-d",
                "2-d",
            ],
        )

    @capture_logging(None)
    def test_close_generator(self, logger):
        @eliot_friendly_generator_function
        def g():
            Message.log(message_type="a")
            try:
                yield
                Message.log(message_type="b")
            finally:
                Message.log(message_type="c")

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            gen = g()
            next(gen)
            gen.close()

        assert_expected_action_tree(self, logger, "the-action", ["a", "yielded", "c"])

    @capture_logging(None)
    def test_nested_generators(self, logger):
        @eliot_friendly_generator_function
        def g(recurse):
            with start_action(action_type="a-recurse={}".format(recurse)):
                Message.log(message_type="m-recurse={}".format(recurse))
                if recurse:
                    set(g(False))
                else:
                    yield

        g.debug = True  # output yielded messages

        with start_action(action_type="the-action"):
            set(g(True))

        assert_expected_action_tree(
            self,
            logger,
            "the-action",
            [
                {
                    "a-recurse=True": [
                        "m-recurse=True",
                        {"a-recurse=False": ["m-recurse=False", "yielded"]},
                    ]
                }
            ],
        )



"""
Utilities to aid unit testing L{eliot} and code that uses it.
"""

from __future__ import unicode_literals

from unittest import SkipTest
from functools import wraps

from pyrsistent import PClass, field
from six import text_type

from ._action import (
    ACTION_STATUS_FIELD,
    ACTION_TYPE_FIELD,
    STARTED_STATUS,
    FAILED_STATUS,
    SUCCEEDED_STATUS,
)
from ._message import MESSAGE_TYPE_FIELD, TASK_LEVEL_FIELD, TASK_UUID_FIELD
from ._output import MemoryLogger
from . import _output

COMPLETED_STATUSES = (FAILED_STATUS, SUCCEEDED_STATUS)


def issuperset(a, b):
    """
    Use L{assertContainsFields} instead.

    @type a: C{dict}

    @type b: C{dict}

    @return: Boolean indicating whether C{a} has all key/value pairs that C{b}
        does.
    """
    aItems = a.items()
    return all(pair in aItems for pair in b.items())


def assertContainsFields(test, message, fields):
    """
    Assert that the given message contains the given fields.

    @param test: L{unittest.TestCase} being run.

    @param message: C{dict}, the message we are checking.

    @param fields: C{dict}, the fields we expect the message to have.

    @raises AssertionError: If the message doesn't contain the fields.
    """
    messageSubset = dict(
        [(key, value) for key, value in message.items() if key in fields]
    )
    test.assertEqual(messageSubset, fields)


class LoggedAction(PClass):
    """
    An action whose start and finish messages have been logged.

    @ivar startMessage: A C{dict}, the start message contents. Also
        available as C{start_message}.

    @ivar endMessage: A C{dict}, the end message contents (in both success and
        failure cases). Also available as C{end_message}.

    @ivar children: A C{list} of direct child L{LoggedMessage} and
        L{LoggedAction} instances.
    """

    startMessage = field(mandatory=True)
    endMessage = field(mandatory=True)
    children = field(mandatory=True)

    def __new__(cls, startMessage, endMessage, children):
        return PClass.__new__(
            cls, startMessage=startMessage, endMessage=endMessage, children=children
        )

    @property
    def start_message(self):
        return self.startMessage

    @property
    def end_message(self):
        return self.endMessage

    @classmethod
    def fromMessages(klass, uuid, level, messages):
        """
        Given a task uuid and level (identifying an action) and a list of
        dictionaries, create a L{LoggedAction}.

        All child messages and actions will be added as L{LoggedAction} or
        L{LoggedMessage} children. Note that some descendant messages may be
        missing if you end up logging to two or more different ILogger
        providers.

        @param uuid: The uuid of the task (C{unicode}).

        @param level: The C{task_level} of the action's start message,
            e.g. C{"/1/2/1"}.

        @param messages: A list of message C{dict}s.

        @return: L{LoggedAction} constructed from start and finish messages for
            this specific action.

        @raises: L{ValueError} if one or both of the action's messages cannot be
            found.
        """
        startMessage = None
        endMessage = None
        children = []
        levelPrefix = level[:-1]

        for message in messages:
            if message[TASK_UUID_FIELD] != uuid:
                # Different task altogether:
                continue

            messageLevel = message[TASK_LEVEL_FIELD]

            if messageLevel[:-1] == levelPrefix:
                status = message.get(ACTION_STATUS_FIELD)
                if status == STARTED_STATUS:
                    startMessage = message
                elif status in COMPLETED_STATUSES:
                    endMessage = message
                else:
                    # Presumably a message in this action:
                    children.append(LoggedMessage(message))
            elif (
                len(messageLevel) == len(levelPrefix) + 2
                and messageLevel[:-2] == levelPrefix
                and messageLevel[-1] == 1
            ):
                # If start message level is [1], [1, 2, 1] implies first
                # message of a direct child.
                child = klass.fromMessages(uuid, message[TASK_LEVEL_FIELD], messages)
                children.append(child)
        if startMessage is None:
            raise ValueError("Missing start message")
        if endMessage is None:
            raise ValueError(
                "Missing end message of type "
                + message.get(ACTION_TYPE_FIELD, "unknown")
            )
        return klass(startMessage, endMessage, children)

    # PEP 8 variant:
    from_messages = fromMessages

    @classmethod
    def of_type(klass, messages, actionType):
        """
        Find all L{LoggedAction} of the specified type.

        @param messages: A list of message C{dict}s.

        @param actionType: A L{eliot.ActionType}, the type of the actions to
            find, or the type as a C{str}.

        @return: A C{list} of L{LoggedAction}.
        """
        if not isinstance(actionType, text_type):
            actionType = actionType.action_type
        result = []
        for message in messages:
            if (
                message.get(ACTION_TYPE_FIELD) == actionType
                and message[ACTION_STATUS_FIELD] == STARTED_STATUS
            ):
                result.append(
                    klass.fromMessages(
                        message[TASK_UUID_FIELD], message[TASK_LEVEL_FIELD], messages
                    )
                )
        return result

    # Backwards compat:
    ofType = of_type

    def descendants(self):
        """
        Find all descendant L{LoggedAction} or L{LoggedMessage} of this
        instance.

        @return: An iterable of L{LoggedAction} and L{LoggedMessage} instances.
        """
        for child in self.children:
            yield child
            if isinstance(child, LoggedAction):
                for descendant in child.descendants():
                    yield descendant

    @property
    def succeeded(self):
        """
        Indicate whether this action succeeded.

        @return: C{bool} indicating whether the action succeeded.
        """
        return self.endMessage[ACTION_STATUS_FIELD] == SUCCEEDED_STATUS

    def type_tree(self):
        """Return dictionary of all child action and message types.

        Actions become dictionaries that look like
        C{{<action_type>: [<child_message_type>, <child_action_dict>]}}

        @return: C{dict} where key is action type, and value is list of child
            types: either strings for messages, or dicts for actions.
        """
        children = []
        for child in self.children:
            if isinstance(child, LoggedAction):
                children.append(child.type_tree())
            else:
                children.append(child.message[MESSAGE_TYPE_FIELD])
        return {self.startMessage[ACTION_TYPE_FIELD]: children}


class LoggedMessage(PClass):
    """
    A message that has been logged.

    @ivar message: A C{dict}, the message contents.
    """

    message = field(mandatory=True)

    def __new__(cls, message):
        return PClass.__new__(cls, message=message)

    @classmethod
    def of_type(klass, messages, messageType):
        """
        Find all L{LoggedMessage} of the specified type.

        @param messages: A list of message C{dict}s.

        @param messageType: A L{eliot.MessageType}, the type of the messages
            to find, or the type as a L{str}.

        @return: A C{list} of L{LoggedMessage}.
        """
        result = []
        if not isinstance(messageType, text_type):
            messageType = messageType.message_type
        for message in messages:
            if message.get(MESSAGE_TYPE_FIELD) == messageType:
                result.append(klass(message))
        return result

    # Backwards compat:
    ofType = of_type


class UnflushedTracebacks(Exception):
    """
    The L{MemoryLogger} had some tracebacks logged which were not flushed.

    This means either your code has a bug and logged an unexpected
    traceback. If you expected the traceback then you will need to flush it
    using L{MemoryLogger.flushTracebacks}.
    """


def check_for_errors(logger):
    """
    Raise exception if logger has unflushed tracebacks or validation errors.

    @param logger: A L{MemoryLogger}.

    @raise L{UnflushedTracebacks}: If any tracebacks were unflushed.
    """
    # Check for unexpected tracebacks first, since that indicates business
    # logic errors:
    if logger.tracebackMessages:
        raise UnflushedTracebacks(logger.tracebackMessages)
    # If those are fine, validate the logging:
    logger.validate()


def swap_logger(logger):
    """Swap out the global logging sink.

    @param logger: An C{ILogger}.

    @return: The current C{ILogger}.
    """
    previous_logger = _output._DEFAULT_LOGGER
    _output._DEFAULT_LOGGER = logger
    return previous_logger


def validateLogging(assertion, *assertionArgs, **assertionKwargs):
    """
    Decorator factory for L{unittest.TestCase} methods to add logging
    validation.

    1. The decorated test method gets a C{logger} keyword argument, a
       L{MemoryLogger}.
    2. All messages logged to this logger will be validated at the end of
       the test.
    3. Any unflushed logged tracebacks will cause the test to fail.

    For example:

        from unittest import TestCase
        from eliot.testing import assertContainsFields, validateLogging

        class MyTests(TestCase):
            def assertFooLogging(self, logger):
                assertContainsFields(self, logger.messages[0], {"key": 123})


    @param assertion: A callable that will be called with the
       L{unittest.TestCase} instance, the logger and C{assertionArgs} and
       C{assertionKwargs} once the actual test has run, allowing for extra
       logging-related assertions on the effects of the test. Use L{None} if you
       want the cleanup assertions registered but no custom assertions.

    @param assertionArgs: Additional positional arguments to pass to
        C{assertion}.

    @param assertionKwargs: Additional keyword arguments to pass to
        C{assertion}.
    """

    def decorator(function):
        @wraps(function)
        def wrapper(self, *args, **kwargs):
            skipped = False

            kwargs["logger"] = logger = MemoryLogger()
            self.addCleanup(check_for_errors, logger)
            # TestCase runs cleanups in reverse order, and we want this to
            # run *before* tracebacks are checked:
            if assertion is not None:
                self.addCleanup(
                    lambda: skipped
                    or assertion(self, logger, *assertionArgs, **assertionKwargs)
                )
            try:
                return function(self, *args, **kwargs)
            except SkipTest:
                skipped = True
                raise

        return wrapper

    return decorator


# PEP 8 variant:
validate_logging = validateLogging


def capture_logging(assertion, *assertionArgs, **assertionKwargs):
    """
    Capture and validate all logging that doesn't specify a L{Logger}.

    See L{validate_logging} for details on the rest of its behavior.
    """

    def decorator(function):
        @validate_logging(assertion, *assertionArgs, **assertionKwargs)
        @wraps(function)
        def wrapper(self, *args, **kwargs):
            logger = kwargs["logger"]
            previous_logger = swap_logger(logger)

            def cleanup():
                swap_logger(previous_logger)

            self.addCleanup(cleanup)
            return function(self, *args, **kwargs)

        return wrapper

    return decorator


def assertHasMessage(testCase, logger, messageType, fields=None):
    """
    Assert that the given logger has a message of the given type, and the first
    message found of this type has the given fields.

    This can be used as the assertion function passed to L{validateLogging} or
    as part of a unit test.

    @param testCase: L{unittest.TestCase} instance.

    @param logger: L{eliot.MemoryLogger} whose messages will be checked.

    @param messageType: L{eliot.MessageType} indicating which message we're
        looking for.

    @param fields: The first message of the given type found must have a
        superset of the given C{dict} as its fields. If C{None} then fields are
        not checked.

    @return: The first found L{LoggedMessage} of the given type, if field
        validation succeeded.

    @raises AssertionError: No message was found, or the fields were not
        superset of given fields.
    """
    if fields is None:
        fields = {}
    messages = LoggedMessage.ofType(logger.messages, messageType)
    testCase.assertTrue(messages, "No messages of type %s" % (messageType,))
    loggedMessage = messages[0]
    assertContainsFields(testCase, loggedMessage.message, fields)
    return loggedMessage


def assertHasAction(
    testCase, logger, actionType, succeeded, startFields=None, endFields=None
):
    """
    Assert that the given logger has an action of the given type, and the first
    action found of this type has the given fields and success status.

    This can be used as the assertion function passed to L{validateLogging} or
    as part of a unit test.

    @param testCase: L{unittest.TestCase} instance.

    @param logger: L{eliot.MemoryLogger} whose messages will be checked.

    @param actionType: L{eliot.ActionType} or C{str} indicating which message
        we're looking for.

    @param succeeded: Expected success status of the action, a C{bool}.

    @param startFields: The first action of the given type found must have a
        superset of the given C{dict} as its start fields.  If C{None} then
        fields are not checked.

    @param endFields: The first action of the given type found must have a
        superset of the given C{dict} as its end fields.  If C{None} then
        fields are not checked.

    @return: The first found L{LoggedAction} of the given type, if field
        validation succeeded.

    @raises AssertionError: No action was found, or the fields were not
        superset of given fields.
    """
    if startFields is None:
        startFields = {}
    if endFields is None:
        endFields = {}
    actions = LoggedAction.ofType(logger.messages, actionType)
    testCase.assertTrue(actions, "No actions of type %s" % (actionType,))
    action = actions[0]
    testCase.assertEqual(action.succeeded, succeeded)
    assertContainsFields(testCase, action.startMessage, startFields)
    assertContainsFields(testCase, action.endMessage, endFields)
    return action



"""
Support for maintaining an action context across generator suspension.
"""

from __future__ import unicode_literals, absolute_import

from sys import exc_info
from functools import wraps
from contextlib import contextmanager
from contextvars import copy_context
from weakref import WeakKeyDictionary

from . import log_message


class _GeneratorContext(object):
    """Generator sub-context for C{_ExecutionContext}."""

    def __init__(self, execution_context):
        self._execution_context = execution_context
        self._contexts = WeakKeyDictionary()
        self._current_generator = None

    def init_stack(self, generator):
        """Create a new stack for the given generator."""
        self._contexts[generator] = copy_context()

    @contextmanager
    def in_generator(self, generator):
        """Context manager: set the given generator as the current generator."""
        previous_generator = self._current_generator
        try:
            self._current_generator = generator
            yield
        finally:
            self._current_generator = previous_generator


class GeneratorSupportNotEnabled(Exception):
    """
    An attempt was made to use a decorated generator without first turning on
    the generator context manager.
    """


def eliot_friendly_generator_function(original):
    """
    Decorate a generator function so that the Eliot action context is
    preserved across ``yield`` expressions.
    """

    @wraps(original)
    def wrapper(*a, **kw):
        # Keep track of whether the next value to deliver to the generator is
        # a non-exception or an exception.
        ok = True

        # Keep track of the next value to deliver to the generator.
        value_in = None

        # Create the generator with a call to the generator function.  This
        # happens with whatever Eliot action context happens to be active,
        # which is fine and correct and also irrelevant because no code in the
        # generator function can run until we call send or throw on it.
        gen = original(*a, **kw)

        # Initialize the per-generator context to a copy of the current context.
        context = copy_context()
        while True:
            try:
                # Whichever way we invoke the generator, we will do it
                # with the Eliot action context stack we've saved for it.
                # Then the context manager will re-save it and restore the
                # "outside" stack for us.
                #
                # Regarding the support of Twisted's inlineCallbacks-like
                # functionality (see eliot.twisted.inline_callbacks):
                #
                # The invocation may raise the inlineCallbacks internal
                # control flow exception _DefGen_Return.  It is not wrong to
                # just let that propagate upwards here but inlineCallbacks
                # does think it is wrong.  The behavior triggers a
                # DeprecationWarning to try to get us to fix our code.  We
                # could explicitly handle and re-raise the _DefGen_Return but
                # only at the expense of depending on a private Twisted API.
                # For now, I'm opting to try to encourage Twisted to fix the
                # situation (or at least not worsen it):
                # https://twistedmatrix.com/trac/ticket/9590
                #
                # Alternatively, _DefGen_Return is only required on Python 2.
                # When Python 2 support is dropped, this concern can be
                # eliminated by always using `return value` instead of
                # `returnValue(value)` (and adding the necessary logic to the
                # StopIteration handler below).
                def go():
                    if ok:
                        value_out = gen.send(value_in)
                    else:
                        value_out = gen.throw(*value_in)
                    # We have obtained a value from the generator.  In
                    # giving it to us, it has given up control.  Note this
                    # fact here.  Importantly, this is within the
                    # generator's action context so that we get a good
                    # indication of where the yield occurred.
                    #
                    # This is noisy, enable only for debugging:
                    if wrapper.debug:
                        log_message(message_type="yielded")
                    return value_out

                value_out = context.run(go)
            except StopIteration:
                # When the generator raises this, it is signaling
                # completion.  Leave the loop.
                break
            else:
                try:
                    # Pass the generator's result along to whoever is
                    # driving.  Capture the result as the next value to
                    # send inward.
                    value_in = yield value_out
                except:
                    # Or capture the exception if that's the flavor of the
                    # next value.  This could possibly include GeneratorExit
                    # which turns out to be just fine because throwing it into
                    # the inner generator effectively propagates the close
                    # (and with the right context!) just as you would want.
                    # True, the GeneratorExit does get re-throwing out of the
                    # gen.throw call and hits _the_generator_context's
                    # contextmanager.  But @contextmanager extremely
                    # conveniently eats it for us!  Thanks, @contextmanager!
                    ok = False
                    value_in = exc_info()
                else:
                    ok = True

    wrapper.debug = False
    return wrapper
